# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import os
import tempfile
from unittest import mock

import cf_units
import numpy as np

import iris.coords
import iris.coord_systems
import iris.fileformats.pp
from iris.fileformats.pp import PPField3
import iris.tests.pp as pp
import iris.util
import iris.tests.stock as stock


def itab_callback(cube, field, filename):
    cube.add_aux_coord(
        iris.coords.AuxCoord(
            [field.lbrel], long_name="MOUMHeaderReleaseNumber", units="no_unit"
        )
    )
    cube.add_aux_coord(
        iris.coords.AuxCoord(
            [field.lbexp], long_name="ExperimentNumber(ITAB)", units="no_unit"
        )
    )


@tests.skip_data
class TestPPSave(tests.IrisTest, pp.PPTest):
    def test_no_forecast_time(self):
        cube = stock.lat_lon_cube()
        coord = iris.coords.DimCoord(
            np.array([24], dtype=np.int64),
            standard_name="time",
            units="hours since epoch",
        )
        cube.add_aux_coord(coord)
        self.assertCML(cube, ["cube_to_pp", "no_forecast_time.cml"])

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "no_forecast_time.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cube
        ) as temp_pp_path:
            iris.save(cube, temp_pp_path)

    def test_no_forecast_period(self):
        cube = stock.lat_lon_cube()
        # Add a bounded scalar time coord and a forecast_reference_time.
        time_coord = iris.coords.DimCoord(
            10.958333,
            standard_name="time",
            units="days since 2013-05-10 12:00",
            bounds=[10.916667, 11.0],
        )
        cube.add_aux_coord(time_coord)
        forecast_reference_time = iris.coords.DimCoord(
            2.0,
            standard_name="forecast_reference_time",
            units="weeks since 2013-05-07",
        )
        cube.add_aux_coord(forecast_reference_time)

        self.assertCML(cube, ["cube_to_pp", "no_forecast_period.cml"])
        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "no_forecast_period.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cube
        ) as temp_pp_path:
            iris.save(cube, temp_pp_path)

    def test_pp_save_rules(self):
        # Test pp save rules without user rules.

        # read
        in_filename = tests.get_data_path(("PP", "simple_pp", "global.pp"))
        cubes = iris.load(in_filename, callback=itab_callback)

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "simple.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cubes
        ) as temp_pp_path:
            iris.save(cubes, temp_pp_path)

    def test_pp_append_singles(self):
        # Test pp append saving - single cubes.

        # load 2 arrays of >2D cubes
        cube = stock.simple_pp()

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "append_single.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=[cube, cube]
        ) as temp_pp_path:
            iris.save(cube, temp_pp_path)  # Create file
            iris.save(cube, temp_pp_path, append=True)  # Append to file

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "replace_single.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cube
        ) as temp_pp_path:
            iris.save(cube, temp_pp_path)  # Create file
            iris.save(cube, temp_pp_path)  # Replace file

    def test_pp_append_lists(self):
        # Test PP append saving - lists of cubes.
        # For each of the first four time-steps in the 4D cube,
        # pull out the bottom two levels.
        cube_4d = stock.realistic_4d()
        cubes = [cube_4d[i, :2, :, :] for i in range(4)]

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "append_multi.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cubes
        ) as temp_pp_path:
            iris.save(cubes[:2], temp_pp_path)
            iris.save(cubes[2:], temp_pp_path, append=True)

        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "replace_multi.txt")
        )
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cubes[2:]
        ) as temp_pp_path:
            iris.save(cubes[:2], temp_pp_path)
            iris.save(cubes[2:], temp_pp_path)

    def add_coords_to_cube_and_test(self, coord1, coord2):
        # a wrapper for creating arbitrary 2d cross-sections and run pp-saving tests
        dataarray = np.arange(16, dtype=">f4").reshape(4, 4)
        cm = iris.cube.Cube(data=dataarray)

        cm.add_dim_coord(coord1, 0)
        cm.add_dim_coord(coord2, 1)

        # TODO: This is the desired line of code...
        # reference_txt_path = tests.get_result_path(('cube_to_pp', '%s.%s.pp.txt' % (coord1.name(), coord2.name())))
        # ...but this is required during the CF change, to maintain the original filename.
        coord1_name = coord1.name().replace("air_", "")
        coord2_name = coord2.name().replace("air_", "")
        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "%s.%s.pp.txt" % (coord1_name, coord2_name))
        )

        # test with name
        with self.cube_save_test(
            reference_txt_path,
            reference_cubes=cm,
            field_coords=[coord1.name(), coord2.name()],
        ) as temp_pp_path:
            iris.save(
                cm, temp_pp_path, field_coords=[coord1.name(), coord2.name()]
            )
        # test with coord
        with self.cube_save_test(
            reference_txt_path,
            reference_cubes=cm,
            field_coords=[coord1, coord2],
        ) as temp_pp_path:
            iris.save(cm, temp_pp_path, field_coords=[coord1, coord2])

    def test_non_standard_cross_sections(self):
        # ticket #1037, the five variants being dealt with are
        #    'pressure.latitude',
        #    'depth.latitude',
        #    'eta.latitude',
        #    'pressure.time',
        #    'depth.time',

        f = FakePPEnvironment()

        self.add_coords_to_cube_and_test(
            iris.coords.DimCoord(
                f.z, long_name="air_pressure", units="hPa", bounds=f.z_bounds
            ),
            iris.coords.DimCoord(
                f.y,
                standard_name="latitude",
                units="degrees",
                bounds=f.y_bounds,
                coord_system=f.geog_cs(),
            ),
        )

        self.add_coords_to_cube_and_test(
            iris.coords.DimCoord(
                f.z, long_name="depth", units="m", bounds=f.z_bounds
            ),
            iris.coords.DimCoord(
                f.y,
                standard_name="latitude",
                units="degrees",
                bounds=f.y_bounds,
                coord_system=f.geog_cs(),
            ),
        )

        self.add_coords_to_cube_and_test(
            iris.coords.DimCoord(
                f.z, long_name="eta", units="1", bounds=f.z_bounds
            ),
            iris.coords.DimCoord(
                f.y,
                standard_name="latitude",
                units="degrees",
                bounds=f.y_bounds,
                coord_system=f.geog_cs(),
            ),
        )

        self.add_coords_to_cube_and_test(
            iris.coords.DimCoord(
                f.z, long_name="air_pressure", units="hPa", bounds=f.z_bounds
            ),
            iris.coords.DimCoord(
                f.y,
                standard_name="time",
                units=cf_units.Unit(
                    "days since 0000-01-01 00:00:00",
                    calendar=cf_units.CALENDAR_360_DAY,
                ),
                bounds=f.y_bounds,
            ),
        )

        self.add_coords_to_cube_and_test(
            iris.coords.DimCoord(
                f.z, standard_name="depth", units="m", bounds=f.z_bounds
            ),
            iris.coords.DimCoord(
                f.y,
                standard_name="time",
                units=cf_units.Unit(
                    "days since 0000-01-01 00:00:00",
                    calendar=cf_units.CALENDAR_360_DAY,
                ),
                bounds=f.y_bounds,
            ),
        )

    def test_365_calendar_export(self):
        # test for 365 day calendar export
        cube = stock.simple_pp()
        new_unit = cf_units.Unit(
            "hours since 1970-01-01 00:00:00",
            calendar=cf_units.CALENDAR_365_DAY,
        )
        cube.coord("time").units = new_unit
        # Add an extra "fill_value" property, as used by the save rules.
        cube.fill_value = None
        pp_field = mock.MagicMock(spec=PPField3)
        iris.fileformats.pp_save_rules.verify(cube, pp_field)
        self.assertEqual(pp_field.lbtim.ic, 4)


class FakePPEnvironment:
    """fake a minimal PP environment for use in cross-section coords, as in PP save rules"""

    y = [1, 2, 3, 4]
    z = [111, 222, 333, 444]
    y_bounds = [[0.9, 1.1], [1.9, 2.1], [2.9, 3.1], [3.9, 4.1]]
    z_bounds = [[110.9, 111.1], [221.9, 222.1], [332.9, 333.1], [443.9, 444.1]]

    def geog_cs(self):
        """Return a GeogCS for this PPField.

        Returns:
            A GeogCS with the appropriate earth shape, meridian and pole position.
        """
        return iris.coord_systems.GeogCS(6371229.0)


class TestPPSaveRules(tests.IrisTest, pp.PPTest):
    def test_default_coord_system(self):
        GeogCS = iris.coord_systems.GeogCS
        cube = iris.tests.stock.lat_lon_cube()
        reference_txt_path = tests.get_result_path(
            ("cube_to_pp", "default_coord_system.txt")
        )
        # Remove all coordinate systems.
        for coord in cube.coords():
            coord.coord_system = None
        # Ensure no coordinate systems available.
        self.assertIsNone(cube.coord_system(GeogCS))
        self.assertIsNone(cube.coord_system(None))
        with self.cube_save_test(
            reference_txt_path, reference_cubes=cube
        ) as temp_pp_path:
            # Save cube to PP with no coordinate system.
            iris.save(cube, temp_pp_path)
            pp_cube = iris.load_cube(temp_pp_path)
            # Ensure saved cube has the default coordinate system.
            self.assertIsInstance(
                pp_cube.coord_system(GeogCS), iris.coord_systems.GeogCS
            )
            self.assertIsNotNone(pp_cube.coord_system(None))
            self.assertIsInstance(
                pp_cube.coord_system(None), iris.coord_systems.GeogCS
            )
            self.assertIsNotNone(pp_cube.coord_system())
            self.assertIsInstance(
                pp_cube.coord_system(), iris.coord_systems.GeogCS
            )

    def lbproc_from_pp(self, filename):
        # Gets the lbproc field from the ppfile
        pp_file = iris.fileformats.pp.load(filename)
        field = next(pp_file)
        return field.lbproc

    def test_pp_save_rules(self):
        # Test single process flags
        for _, process_desc in iris.fileformats.pp.LBPROC_PAIRS[1:]:
            # Get basic cube and set process flag manually
            ll_cube = stock.lat_lon_cube()
            ll_cube.attributes["ukmo__process_flags"] = (process_desc,)

            # Save cube to pp
            temp_filename = iris.util.create_temp_filename(".pp")
            iris.save(ll_cube, temp_filename)

            # Check the lbproc is what we expect
            self.assertEqual(
                self.lbproc_from_pp(temp_filename),
                iris.fileformats.pp.lbproc_map[process_desc],
            )

            os.remove(temp_filename)

        # Test mutiple process flags
        multiple_bit_values = ((128, 64), (4096, 1024), (8192, 1024))

        # Maps lbproc value to the process flags that should be created
        multiple_map = {
            sum(bits): [iris.fileformats.pp.lbproc_map[bit] for bit in bits]
            for bits in multiple_bit_values
        }

        for lbproc, descriptions in multiple_map.items():
            ll_cube = stock.lat_lon_cube()
            ll_cube.attributes["ukmo__process_flags"] = descriptions

            # Save cube to pp
            temp_filename = iris.util.create_temp_filename(".pp")
            iris.save(ll_cube, temp_filename)

            # Check the lbproc is what we expect
            self.assertEqual(self.lbproc_from_pp(temp_filename), lbproc)

            os.remove(temp_filename)

    @tests.skip_data
    def test_lbvc(self):
        cube = stock.realistic_4d_no_derived()[0, :4, ...]

        v_coord = iris.coords.DimCoord(
            standard_name="depth", units="m", points=[-5, -10, -15, -20]
        )

        cube.remove_coord("level_height")
        cube.remove_coord("sigma")
        cube.remove_coord("surface_altitude")
        cube.add_aux_coord(v_coord, 0)

        expected = ([2, 1, -5.0], [2, 2, -10.0], [2, 3, -15.0], [2, 4, -20.0])

        for field, (lbvc, lblev, blev) in zip(
            fields_from_cube(cube), expected
        ):
            self.assertEqual(field.lbvc, lbvc)
            self.assertEqual(field.lblev, lblev)
            self.assertEqual(field.blev, blev)


def fields_from_cube(cubes):
    """
    Return an iterator of PP fields generated from saving the given cube(s)
    to a temporary file, and then subsequently loading them again

    """
    with tempfile.NamedTemporaryFile("w+b", suffix=".pp") as tmp_file:
        fh = tmp_file
        iris.save(cubes, fh, saver="pp")

        # make sure the fh is written to disk, and move it back to the
        # start of the file
        fh.flush()
        os.fsync(fh)
        fh.seek(0)

        # load in the saved pp fields and check the appropriate metadata
        for field in iris.fileformats.pp.load(tmp_file.name):
            yield field


if __name__ == "__main__":
    tests.main()



"""
Load a Time Series of Data From the NEMO Model
==============================================

This example demonstrates how to load multiple files containing data output by
the NEMO model and combine them into a time series in a single cube. The
different time dimensions in these files can prevent Iris from concatenating
them without the intervention shown here.
"""

from __future__ import unicode_literals

import matplotlib.pyplot as plt

import iris
import iris.plot as iplt
import iris.quickplot as qplt
from iris.util import promote_aux_coord_to_dim_coord


def main():
    # Load the three files of sample NEMO data.
    fname = iris.sample_data_path("NEMO/nemo_1m_*.nc")
    cubes = iris.load(fname)

    # Some attributes are unique to each file and must be blanked
    # to allow concatenation.
    differing_attrs = ["file_name", "name", "timeStamp", "TimeStamp"]
    for cube in cubes:
        for attribute in differing_attrs:
            cube.attributes[attribute] = ""

    # The cubes still cannot be concatenated because their time dimension is
    # time_counter rather than time. time needs to be promoted to allow
    # concatenation.
    for cube in cubes:
        promote_aux_coord_to_dim_coord(cube, "time")

    # The cubes can now be concatenated into a single time series.
    cube = cubes.concatenate_cube()

    # Generate a time series plot of a single point
    plt.figure()
    y_point_index = 100
    x_point_index = 100
    qplt.plot(cube[:, y_point_index, x_point_index], "o-")

    # Include the point's position in the plot's title
    lat_point = cube.coord("latitude").points[y_point_index, x_point_index]
    lat_string = "{:.3f}\u00B0 {}".format(
        abs(lat_point), "N" if lat_point > 0.0 else "S"
    )
    lon_point = cube.coord("longitude").points[y_point_index, x_point_index]
    lon_string = "{:.3f}\u00B0 {}".format(
        abs(lon_point), "E" if lon_point > 0.0 else "W"
    )
    plt.title(
        "{} at {} {}".format(
            cube.long_name.capitalize(), lat_string, lon_string
        )
    )

    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Experimental module for importing/exporting raster data from Iris cubes using
the GDAL library.

See also: `GDAL - Geospatial Data Abstraction Library <http://www.gdal.org>`_.

TODO: If this module graduates from experimental the (optional) GDAL
      dependency should be added to INSTALL

"""

import numpy as np
import numpy.ma as ma
from osgeo import gdal, osr

import cf_units

import iris
import iris.coord_systems


_GDAL_DATATYPES = {
    "i2": gdal.GDT_Int16,
    "i4": gdal.GDT_Int32,
    "u1": gdal.GDT_Byte,
    "u2": gdal.GDT_UInt16,
    "u4": gdal.GDT_UInt32,
    "f4": gdal.GDT_Float32,
    "f8": gdal.GDT_Float64,
}


def _gdal_write_array(
    x_min, x_step, y_max, y_step, coord_system, data, fname, ftype
):
    """
    Use GDAL WriteArray to export data as a 32-bit raster image.
    Requires the array data to be of the form: North-at-top
    and West-on-left.

    Args:
        * x_min: Minimum X coordinate bounds value.
        * x_step: Change in X coordinate per cell.
        * y_max: Maximum Y coordinate bounds value.
        * y_step: Change in Y coordinate per cell.
        * coord_system (iris.coord_systems.CoordSystem):
            Coordinate system for X and Y.
        * data (numpy.ndarray): 2d array of values to export
        * fname (string): Output file name.
        * ftype (string): Export file type.

    .. note::

        Projection information is currently not written to the output.

    """
    byte_order = data.dtype.str[0]
    format = data.dtype.str[1:]
    dtype = _GDAL_DATATYPES.get(format)
    if dtype is None:
        raise ValueError("Unsupported data type: {}".format(data.dtype))

    driver = gdal.GetDriverByName(ftype)
    gdal_dataset = driver.Create(fname, data.shape[1], data.shape[0], 1, dtype)

    # Where possible, set the projection.
    if coord_system is not None:
        srs = osr.SpatialReference()
        proj4_defn = coord_system.as_cartopy_crs().proj4_init
        # GDAL can't cope with "+proj=lonlat" which Cartopy produces.
        proj4_defn = proj4_defn.replace("lonlat", "longlat")
        if srs.ImportFromProj4(proj4_defn):
            msg = "Unsupported coordinate system: {}".format(coord_system)
            raise ValueError(msg)
        gdal_dataset.SetProjection(srs.ExportToWkt())

    # Set the affine transformation coefficients.
    padf_transform = (x_min, x_step, 0.0, y_max, 0.0, y_step)
    gdal_dataset.SetGeoTransform(padf_transform)

    band = gdal_dataset.GetRasterBand(1)
    if ma.isMaskedArray(data):
        data = data.copy()
        data[data.mask] = data.fill_value
        band.SetNoDataValue(float(data.fill_value))
    # GeoTIFF always needs little-endian data.
    if byte_order == ">":
        data = data.astype(data.dtype.newbyteorder("<"))
    band.WriteArray(data)


def export_geotiff(cube, fname):
    """
    Writes cube data to raster file format as a PixelIsArea GeoTiff image.

    Args:
        * cube (Cube): The 2D regularly gridded cube slice to be exported.
                       The cube must have regular, contiguous bounds.
        * fname (string): Output file name.

    .. note::

        For more details on GeoTiff specification and PixelIsArea, see:
        http://www.remotesensing.org/geotiff/spec/geotiff2.5.html#2.5.2.2

    """
    if cube.ndim != 2:
        raise ValueError("The cube must be two dimensional.")

    coord_x = cube.coord(axis="X", dim_coords=True)
    coord_y = cube.coord(axis="Y", dim_coords=True)

    if coord_x.bounds is None or coord_y.bounds is None:
        raise ValueError(
            "Coordinates must have bounds, consider using " "guess_bounds()"
        )

    if (
        coord_x is None
        or coord_y is None
        or coord_x.coord_system != coord_y.coord_system
    ):
        raise ValueError("The X and Y coordinates must share a CoordSystem.")

    xy_step = []
    for coord in [coord_x, coord_y]:
        name = coord.name()
        if coord.nbounds != 2:
            msg = "Coordinate {!r} must have two bounds " "per point.".format(
                name
            )
            raise ValueError(msg)
        if not (
            coord.units == cf_units.Unit("degrees")
            or coord.units.is_convertible("meters")
        ):
            raise ValueError(
                "Coordinate {!r} units must be either degrees or "
                "convertible to meters.".format(name)
            )
        if not coord.is_contiguous():
            raise ValueError(
                "Coordinate {!r} bounds must be " "contiguous.".format(name)
            )
        xy_step.append(np.diff(coord.bounds[0]))
        if not np.allclose(np.diff(coord.bounds), xy_step[-1]):
            msg = "Coordinate {!r} bounds must be regularly " "spaced.".format(
                name
            )
            raise ValueError(msg)

    if coord_x.points[0] > coord_x.points[-1]:
        raise ValueError(
            "Coordinate {!r} x-points must be monotonically"
            "increasing.".format(name)
        )

    data = cube.data

    # Make sure we have a YX data layout.
    if cube.coord_dims(coord_x) == 0:
        data = data.T

    x_step, y_step = xy_step
    if y_step > 0:
        # Flip the data so North is at the top.
        data = data[::-1, :]
        y_step *= -1

    coord_system = coord_x.coord_system
    x_bounds = coord_x.bounds
    if isinstance(coord_system, iris.coord_systems.GeogCS):
        big_indices = np.where(coord_x.points > 180)[0]
        n_big = len(big_indices)
        if n_big:
            data = np.roll(data, n_big, axis=1)
            x_bounds = x_bounds.copy()
            x_bounds[big_indices] -= 360

    x_min = np.min(x_bounds)
    y_max = np.max(coord_y.bounds)
    _gdal_write_array(
        x_min, x_step, y_max, y_step, coord_system, data, fname, "GTiff"
    )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.new_axis`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests
import iris.tests.stock as stock

import copy
import numpy as np
import unittest

import iris
from iris._lazy_data import as_lazy_data

from iris.util import new_axis


class Test(tests.IrisTest):
    def setUp(self):
        self.data = np.array([[1, 2], [1, 2]])
        self.cube = iris.cube.Cube(self.data)
        lat = iris.coords.DimCoord([1, 2], standard_name="latitude")
        lon = iris.coords.DimCoord([1, 2], standard_name="longitude")

        time = iris.coords.DimCoord([1], standard_name="time")
        wibble = iris.coords.AuxCoord([1], long_name="wibble")

        self.cube.add_dim_coord(lat, 0)
        self.cube.add_dim_coord(lon, 1)
        self.cube.add_aux_coord(time, None)
        self.cube.add_aux_coord(wibble, None)

        self.coords = {"lat": lat, "lon": lon, "time": time, "wibble": wibble}

    def _assert_cube_notis(self, cube_a, cube_b):
        for coord_a, coord_b in zip(cube_a.coords(), cube_b.coords()):
            self.assertIsNot(coord_a, coord_b)

        self.assertIsNot(cube_a.metadata, cube_b.metadata)

        for factory_a, factory_b in zip(
            cube_a.aux_factories, cube_b.aux_factories
        ):
            self.assertIsNot(factory_a, factory_b)

    def test_no_coord(self):
        # Providing no coordinate to promote.
        res = new_axis(self.cube)
        com = iris.cube.Cube(self.data[None])
        com.add_dim_coord(self.coords["lat"].copy(), 1)
        com.add_dim_coord(self.coords["lon"].copy(), 2)
        com.add_aux_coord(self.coords["time"].copy(), None)
        com.add_aux_coord(self.coords["wibble"].copy(), None)

        self.assertEqual(res, com)
        self._assert_cube_notis(res, self.cube)

    def test_scalar_dimcoord(self):
        # Providing a scalar coordinate to promote.
        res = new_axis(self.cube, "time")
        com = iris.cube.Cube(self.data[None])
        com.add_dim_coord(self.coords["lat"].copy(), 1)
        com.add_dim_coord(self.coords["lon"].copy(), 2)
        com.add_aux_coord(self.coords["time"].copy(), 0)
        com.add_aux_coord(self.coords["wibble"].copy(), None)

        self.assertEqual(res, com)
        self._assert_cube_notis(res, self.cube)

    def test_scalar_auxcoord(self):
        # Providing a scalar coordinate to promote.
        res = new_axis(self.cube, "wibble")
        com = iris.cube.Cube(self.data[None])
        com.add_dim_coord(self.coords["lat"].copy(), 1)
        com.add_dim_coord(self.coords["lon"].copy(), 2)
        com.add_aux_coord(self.coords["time"].copy(), None)
        com.add_aux_coord(self.coords["wibble"].copy(), 0)

        self.assertEqual(res, com)
        self._assert_cube_notis(res, self.cube)

    def test_maint_factory(self):
        # Ensure that aux factory persists.
        data = np.arange(12, dtype="i8").reshape((3, 4))

        orography = iris.coords.AuxCoord(
            [10, 25, 50, 5], standard_name="surface_altitude", units="m"
        )

        model_level = iris.coords.AuxCoord(
            [2, 1, 0], standard_name="model_level_number"
        )

        level_height = iris.coords.DimCoord(
            [100, 50, 10],
            long_name="level_height",
            units="m",
            attributes={"positive": "up"},
            bounds=[[150, 75], [75, 20], [20, 0]],
        )

        sigma = iris.coords.AuxCoord(
            [0.8, 0.9, 0.95],
            long_name="sigma",
            bounds=[[0.7, 0.85], [0.85, 0.97], [0.97, 1.0]],
        )

        hybrid_height = iris.aux_factory.HybridHeightFactory(
            level_height, sigma, orography
        )

        cube = iris.cube.Cube(
            data,
            standard_name="air_temperature",
            units="K",
            dim_coords_and_dims=[(level_height, 0)],
            aux_coords_and_dims=[(orography, 1), (model_level, 0), (sigma, 0)],
            aux_factories=[hybrid_height],
        )

        com = iris.cube.Cube(
            data[None],
            standard_name="air_temperature",
            units="K",
            dim_coords_and_dims=[(copy.copy(level_height), 1)],
            aux_coords_and_dims=[
                (copy.copy(orography), 2),
                (copy.copy(model_level), 1),
                (copy.copy(sigma), 1),
            ],
            aux_factories=[copy.copy(hybrid_height)],
        )
        res = new_axis(cube)

        self.assertEqual(res, com)
        self._assert_cube_notis(res, cube)

        # Check that factory dependencies are actual coords within the cube.
        # Addresses a former bug : https://github.com/SciTools/iris/pull/3263
        (factory,) = list(res.aux_factories)
        deps = factory.dependencies
        for dep_name, dep_coord in deps.items():
            coord_name = dep_coord.name()
            msg = (
                "Factory dependency {!r} is a coord named {!r}, "
                "but it is *not* the coord of that name in the new cube."
            )
            self.assertIs(
                dep_coord,
                res.coord(coord_name),
                msg.format(dep_name, coord_name),
            )

    def test_lazy_data(self):
        cube = iris.cube.Cube(as_lazy_data(self.data))
        cube.add_aux_coord(iris.coords.DimCoord([1], standard_name="time"))
        res = new_axis(cube, "time")
        self.assertTrue(cube.has_lazy_data())
        self.assertTrue(res.has_lazy_data())
        self.assertEqual(res.shape, (1,) + cube.shape)

    def test_masked_unit_array(self):
        cube = stock.simple_3d_mask()
        test_cube = cube[0, 0, 0]
        test_cube = new_axis(test_cube, "longitude")
        test_cube = new_axis(test_cube, "latitude")
        data_shape = test_cube.data.shape
        mask_shape = test_cube.data.mask.shape
        self.assertEqual(data_shape, mask_shape)


if __name__ == "__main__":
    unittest.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.MAX` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.analysis import MAX
from iris.cube import Cube
from iris.coords import DimCoord
from iris._lazy_data import as_lazy_data, is_lazy_data


class Test_basics(tests.IrisTest):
    def setUp(self):
        data = np.array([1, 2, 3, 4, 5])
        coord = DimCoord([6, 7, 8, 9, 10], long_name="foo")
        self.cube = Cube(data)
        self.cube.add_dim_coord(coord, 0)
        self.lazy_cube = Cube(as_lazy_data(data))
        self.lazy_cube.add_dim_coord(coord, 0)

    def test_name(self):
        self.assertEqual(MAX.name(), "maximum")

    def test_collapse(self):
        data = MAX.aggregate(self.cube.data, axis=0)
        self.assertArrayEqual(data, [5])

    def test_lazy(self):
        lazy_data = MAX.lazy_aggregate(self.lazy_cube.lazy_data(), axis=0)
        self.assertTrue(is_lazy_data(lazy_data))

    def test_lazy_collapse(self):
        lazy_data = MAX.lazy_aggregate(self.lazy_cube.lazy_data(), axis=0)
        self.assertArrayEqual(lazy_data.compute(), [5])


class Test_masked(tests.IrisTest):
    def setUp(self):
        self.cube = Cube(ma.masked_greater([1, 2, 3, 4, 5], 3))
        self.cube.add_dim_coord(DimCoord([6, 7, 8, 9, 10], long_name="foo"), 0)

    def test_ma(self):
        data = MAX.aggregate(self.cube.data, axis=0)
        self.assertArrayEqual(data, [3])


class Test_lazy_masked(tests.IrisTest):
    def setUp(self):
        masked_data = ma.masked_greater([1, 2, 3, 4, 5], 3)
        self.cube = Cube(as_lazy_data(masked_data))
        self.cube.add_dim_coord(DimCoord([6, 7, 8, 9, 10], long_name="foo"), 0)

    def test_lazy_ma(self):
        lazy_data = MAX.lazy_aggregate(self.cube.lazy_data(), axis=0)
        self.assertTrue(is_lazy_data(lazy_data))
        self.assertArrayEqual(lazy_data.compute(), [3])


class Test_aggregate_shape(tests.IrisTest):
    def test(self):
        shape = ()
        kwargs = dict()
        self.assertTupleEqual(MAX.aggregate_shape(**kwargs), shape)
        kwargs = dict(wibble="wobble")
        self.assertTupleEqual(MAX.aggregate_shape(**kwargs), shape)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :mod:`iris.fileformats` package."""

import iris.tests as tests


class TestField(tests.IrisTest):
    def _test_for_coord(
        self, field, convert, coord_predicate, expected_points, expected_bounds
    ):
        (
            factories,
            references,
            standard_name,
            long_name,
            units,
            attributes,
            cell_methods,
            dim_coords_and_dims,
            aux_coords_and_dims,
        ) = convert(field)

        # Check for one and only one matching coordinate.
        coords_and_dims = dim_coords_and_dims + aux_coords_and_dims
        matching_coords = [
            coord for coord, _ in coords_and_dims if coord_predicate(coord)
        ]
        self.assertEqual(len(matching_coords), 1, str(matching_coords))
        coord = matching_coords[0]

        # Check points and bounds.
        if expected_points is not None:
            self.assertArrayEqual(coord.points, expected_points)

        if expected_bounds is None:
            self.assertIsNone(coord.bounds)
        else:
            self.assertArrayEqual(coord.bounds, expected_bounds)

    def assertCoordsAndDimsListsMatch(
        self, coords_and_dims_got, coords_and_dims_expected
    ):
        """
        Check that coords_and_dims lists are equivalent.

        The arguments are lists of pairs of (coordinate, dimensions).
        The elements are compared one-to-one, by coordinate name (so the order
        of the lists is _not_ significant).
        It also checks that the coordinate types (DimCoord/AuxCoord) match.

        """

        def sorted_by_coordname(list):
            return sorted(list, key=lambda item: item[0].name())

        coords_and_dims_got = sorted_by_coordname(coords_and_dims_got)
        coords_and_dims_expected = sorted_by_coordname(
            coords_and_dims_expected
        )
        self.assertEqual(coords_and_dims_got, coords_and_dims_expected)
        # Also check coordinate type equivalences (as Coord.__eq__ does not).
        self.assertEqual(
            [type(coord) for coord, dims in coords_and_dims_got],
            [type(coord) for coord, dims in coords_and_dims_expected],
        )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coord_systems.TransverseMercator` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coord_systems import TransverseMercator


class Test_init_defaults(tests.IrisTest):
    # NOTE: most of the testing for TransverseMercator is in the legacy test
    # module 'iris.tests.test_coordsystem'.
    # This class *only* tests the defaults for optional constructor args.

    def test_set_optional_args(self):
        # Check that setting the optional (non-ellipse) args works.
        crs = TransverseMercator(
            0,
            50,
            false_easting=100,
            false_northing=-203.7,
            scale_factor_at_central_meridian=1.057,
        )
        self.assertEqualAndKind(crs.false_easting, 100.0)
        self.assertEqualAndKind(crs.false_northing, -203.7)
        self.assertEqualAndKind(crs.scale_factor_at_central_meridian, 1.057)

    def _check_crs_defaults(self, crs):
        # Check for property defaults when no kwargs options were set.
        # NOTE: except ellipsoid, which is done elsewhere.
        self.assertEqualAndKind(crs.false_easting, 0.0)
        self.assertEqualAndKind(crs.false_northing, 0.0)
        self.assertEqualAndKind(crs.scale_factor_at_central_meridian, 1.0)

    def test_no_optional_args(self):
        # Check expected defaults with no optional args.
        crs = TransverseMercator(0, 50)
        self._check_crs_defaults(crs)

    def test_optional_args_None(self):
        # Check expected defaults with optional args=None.
        crs = TransverseMercator(
            0,
            50,
            false_easting=None,
            false_northing=None,
            scale_factor_at_central_meridian=None,
        )
        self._check_crs_defaults(crs)


if __name__ == "__main__":
    tests.main()



"""
Colouring Anomaly Data With Logarithmic Scaling
===============================================

In this example, we need to plot anomaly data where the values have a
"logarithmic" significance  -- i.e. we want to give approximately equal ranges
of colour between data values of, say, 1 and 10 as between 10 and 100.

As the data range also contains zero, that obviously does not suit a simple
logarithmic interpretation.  However, values of less than a certain absolute
magnitude may be considered "not significant", so we put these into a separate
"zero band" which is plotted in white.

To do this, we create a custom value mapping function (normalization) using
the matplotlib Norm class :obj:`matplotlib.colors.SymLogNorm`.
We use this to make a cell-filled pseudocolor plot with a colorbar.

NOTE: By "pseudocolour", we mean that each data point is drawn as a "cell"
region on the plot, coloured according to its data value.
This is provided in Iris by the functions :meth:`iris.plot.pcolor` and
:meth:`iris.plot.pcolormesh`, which call the underlying matplotlib
functions of the same names (i.e., :obj:`matplotlib.pyplot.pcolor`
and :obj:`matplotlib.pyplot.pcolormesh`).
See also: http://en.wikipedia.org/wiki/False_color#Pseudocolor.

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import matplotlib.colors as mcols

import iris
import iris.coord_categorisation
import iris.plot as iplt


def main():
    # Load a sample air temperatures sequence.
    file_path = iris.sample_data_path("E1_north_america.nc")
    temperatures = iris.load_cube(file_path)

    # Create a year-number coordinate from the time information.
    iris.coord_categorisation.add_year(temperatures, "time")

    # Create a sample anomaly field for one chosen year, by extracting that
    # year and subtracting the time mean.
    sample_year = 1982
    year_temperature = temperatures.extract(iris.Constraint(year=sample_year))
    time_mean = temperatures.collapsed("time", iris.analysis.MEAN)
    anomaly = year_temperature - time_mean

    # Construct a plot title string explaining which years are involved.
    years = temperatures.coord("year").points
    plot_title = "Temperature anomaly"
    plot_title += "\n{} differences from {}-{} average.".format(
        sample_year, years[0], years[-1]
    )

    # Define scaling levels for the logarithmic colouring.
    minimum_log_level = 0.1
    maximum_scale_level = 3.0

    # Use a standard colour map which varies blue-white-red.
    # For suitable options, see the 'Diverging colormaps' section in:
    # http://matplotlib.org/stable/gallery/color/colormap_reference.html
    anom_cmap = "bwr"

    # Create a 'logarithmic' data normalization.
    anom_norm = mcols.SymLogNorm(
        linthresh=minimum_log_level,
        linscale=0.01,
        vmin=-maximum_scale_level,
        vmax=maximum_scale_level,
    )
    # Setting "linthresh=minimum_log_level" makes its non-logarithmic
    # data range equal to our 'zero band'.
    # Setting "linscale=0.01" maps the whole zero band to the middle colour value
    # (i.e., 0.5), which is the neutral point of a "diverging" style colormap.

    # Create an Axes, specifying the map projection.
    plt.axes(projection=ccrs.LambertConformal())

    # Make a pseudocolour plot using this colour scheme.
    mesh = iplt.pcolormesh(anomaly, cmap=anom_cmap, norm=anom_norm)

    # Add a colourbar, with extensions to show handling of out-of-range values.
    bar = plt.colorbar(mesh, orientation="horizontal", extend="both")

    # Set some suitable fixed "logarithmic" colourbar tick positions.
    tick_levels = [-3, -1, -0.3, 0.0, 0.3, 1, 3]
    bar.set_ticks(tick_levels)

    # Modify the tick labels so that the centre one shows "+/-<minumum-level>".
    tick_levels[3] = r"$\pm${:g}".format(minimum_log_level)
    bar.set_ticklabels(tick_levels)

    # Label the colourbar to show the units.
    bar.set_label("[{}, log scale]".format(anomaly.units))

    # Add coastlines and a title.
    plt.gca().coastlines()
    plt.title(plot_title)

    # Display the result.
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for loading and saving netcdf files."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from contextlib import contextmanager
from itertools import repeat
import os.path
from os.path import join as path_join
import shutil
from subprocess import check_call
import tempfile
from unittest import mock
import warnings

import netCDF4 as nc
import numpy as np
import numpy.ma as ma

import iris
from iris.coords import CellMethod
from iris.cube import Cube, CubeList
from iris.fileformats.netcdf import CF_CONVENTIONS_VERSION
from iris.fileformats.netcdf import Saver
from iris.fileformats.netcdf import UnknownCellMethodWarning
import iris.tests.stock as stock


@tests.skip_data
class TestHybridPressure(tests.IrisTest):
    def setUp(self):
        # Modify stock cube so it is suitable to have a
        # hybrid pressure factory added to it.
        cube = stock.realistic_4d_no_derived()
        cube.coord("surface_altitude").rename("surface_air_pressure")
        cube.coord("surface_air_pressure").units = "Pa"
        cube.coord("level_height").rename("level_pressure")
        cube.coord("level_pressure").units = "Pa"
        # Construct and add hybrid pressure factory.
        factory = iris.aux_factory.HybridPressureFactory(
            cube.coord("level_pressure"),
            cube.coord("sigma"),
            cube.coord("surface_air_pressure"),
        )
        cube.add_aux_factory(factory)
        self.cube = cube

    def test_save(self):
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(self.cube, filename)
            self.assertCDL(filename)

    def test_save_load_loop(self):
        # Tests an issue where the variable names in the formula
        # terms changed to the standard_names instead of the variable names
        # when loading a previously saved cube.
        with self.temp_filename(suffix=".nc") as filename, self.temp_filename(
            suffix=".nc"
        ) as other_filename:
            iris.save(self.cube, filename)
            cube = iris.load_cube(filename, "air_potential_temperature")
            iris.save(cube, other_filename)
            other_cube = iris.load_cube(
                other_filename, "air_potential_temperature"
            )
            self.assertEqual(cube, other_cube)


@tests.skip_data
class TestSaveMultipleAuxFactories(tests.IrisTest):
    def test_hybrid_height_and_pressure(self):
        cube = stock.realistic_4d()
        cube.add_aux_coord(
            iris.coords.DimCoord(
                1200.0, long_name="level_pressure", units="hPa"
            )
        )
        cube.add_aux_coord(
            iris.coords.DimCoord(0.5, long_name="other sigma", units="1")
        )
        cube.add_aux_coord(
            iris.coords.DimCoord(
                1000.0, long_name="surface_air_pressure", units="hPa"
            )
        )
        factory = iris.aux_factory.HybridPressureFactory(
            cube.coord("level_pressure"),
            cube.coord("other sigma"),
            cube.coord("surface_air_pressure"),
        )
        cube.add_aux_factory(factory)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(cube, filename)
            self.assertCDL(filename)

    def test_shared_primary(self):
        cube = stock.realistic_4d()
        factory = iris.aux_factory.HybridHeightFactory(
            cube.coord("level_height"),
            cube.coord("sigma"),
            cube.coord("surface_altitude"),
        )
        factory.rename("another altitude")
        cube.add_aux_factory(factory)
        with self.temp_filename(
            suffix=".nc"
        ) as filename, self.assertRaisesRegex(
            ValueError, "multiple aux factories"
        ):
            iris.save(cube, filename)

    def test_hybrid_height_cubes(self):
        hh1 = stock.simple_4d_with_hybrid_height()
        hh1.attributes["cube"] = "hh1"
        hh2 = stock.simple_4d_with_hybrid_height()
        hh2.attributes["cube"] = "hh2"
        sa = hh2.coord("surface_altitude")
        sa.points = sa.points * 10
        with self.temp_filename(".nc") as fname:
            iris.save([hh1, hh2], fname)
            cubes = iris.load(fname, "air_temperature")
            cubes = sorted(cubes, key=lambda cube: cube.attributes["cube"])
            self.assertCML(cubes)

    def test_hybrid_height_cubes_on_dimension_coordinate(self):
        hh1 = stock.hybrid_height()
        hh2 = stock.hybrid_height()
        sa = hh2.coord("surface_altitude")
        sa.points = sa.points * 10
        emsg = "Unable to create dimensonless vertical coordinate."
        with self.temp_filename(".nc") as fname, self.assertRaisesRegex(
            ValueError, emsg
        ):
            iris.save([hh1, hh2], fname)


class TestUmVersionAttribute(tests.IrisTest):
    def test_single_saves_as_global(self):
        cube = Cube(
            [1.0],
            standard_name="air_temperature",
            units="K",
            attributes={"um_version": "4.3"},
        )
        with self.temp_filename(".nc") as nc_path:
            iris.save(cube, nc_path)
            self.assertCDL(nc_path)

    def test_multiple_same_saves_as_global(self):
        cube_a = Cube(
            [1.0],
            standard_name="air_temperature",
            units="K",
            attributes={"um_version": "4.3"},
        )
        cube_b = Cube(
            [1.0],
            standard_name="air_pressure",
            units="hPa",
            attributes={"um_version": "4.3"},
        )
        with self.temp_filename(".nc") as nc_path:
            iris.save(CubeList([cube_a, cube_b]), nc_path)
            self.assertCDL(nc_path)

    def test_multiple_different_saves_on_variables(self):
        cube_a = Cube(
            [1.0],
            standard_name="air_temperature",
            units="K",
            attributes={"um_version": "4.3"},
        )
        cube_b = Cube(
            [1.0],
            standard_name="air_pressure",
            units="hPa",
            attributes={"um_version": "4.4"},
        )
        with self.temp_filename(".nc") as nc_path:
            iris.save(CubeList([cube_a, cube_b]), nc_path)
            self.assertCDL(nc_path)


@contextmanager
def _patch_site_configuration():
    def cf_patch_conventions(conventions):
        return ", ".join([conventions, "convention1, convention2"])

    def update(config):
        config["cf_profile"] = mock.Mock(name="cf_profile")
        config["cf_patch"] = mock.Mock(name="cf_patch")
        config["cf_patch_conventions"] = cf_patch_conventions

    orig_site_config = iris.site_configuration.copy()
    update(iris.site_configuration)
    yield
    iris.site_configuration = orig_site_config


class TestConventionsAttributes(tests.IrisTest):
    def test_patching_conventions_attribute(self):
        # Ensure that user defined conventions are wiped and those which are
        # saved patched through site_config can be loaded without an exception
        # being raised.
        cube = Cube(
            [1.0],
            standard_name="air_temperature",
            units="K",
            attributes={"Conventions": "some user defined conventions"},
        )

        # Patch the site configuration dictionary.
        with _patch_site_configuration(), self.temp_filename(".nc") as nc_path:
            iris.save(cube, nc_path)
            res = iris.load_cube(nc_path)

        self.assertEqual(
            res.attributes["Conventions"],
            "{}, {}, {}".format(
                CF_CONVENTIONS_VERSION, "convention1", "convention2"
            ),
        )


class TestLazySave(tests.IrisTest):
    @tests.skip_data
    def test_lazy_preserved_save(self):
        fpath = tests.get_data_path(
            ("NetCDF", "label_and_climate", "small_FC_167_mon_19601101.nc")
        )
        acube = iris.load_cube(fpath, "air_temperature")
        self.assertTrue(acube.has_lazy_data())
        with self.temp_filename(".nc") as nc_path:
            with Saver(nc_path, "NETCDF4") as saver:
                saver.write(acube)
        self.assertTrue(acube.has_lazy_data())


@tests.skip_data
class TestCellMeasures(tests.IrisTest):
    def setUp(self):
        self.fname = tests.get_data_path(("NetCDF", "ORCA2", "votemper.nc"))

    def test_load_raw(self):
        (cube,) = iris.load_raw(self.fname)
        self.assertEqual(len(cube.cell_measures()), 1)
        self.assertEqual(cube.cell_measures()[0].measure, "area")

    def test_load(self):
        cube = iris.load_cube(self.fname)
        self.assertEqual(len(cube.cell_measures()), 1)
        self.assertEqual(cube.cell_measures()[0].measure, "area")

    def test_merge_cell_measure_aware(self):
        (cube1,) = iris.load_raw(self.fname)
        (cube2,) = iris.load_raw(self.fname)
        cube2._cell_measures_and_dims[0][0].var_name = "not_areat"
        cubes = CubeList([cube1, cube2]).merge()
        self.assertEqual(len(cubes), 2)

    def test_concatenate_cell_measure_aware(self):
        (cube1,) = iris.load_raw(self.fname)
        cube1 = cube1[:, :, 0, 0]
        cm_and_dims = cube1._cell_measures_and_dims
        (cube2,) = iris.load_raw(self.fname)
        cube2 = cube2[:, :, 0, 0]
        cube2._cell_measures_and_dims[0][0].var_name = "not_areat"
        cube2.coord("time").points = cube2.coord("time").points + 1
        cubes = CubeList([cube1, cube2]).concatenate()
        self.assertEqual(cubes[0]._cell_measures_and_dims, cm_and_dims)
        self.assertEqual(len(cubes), 2)

    def test_concatenate_cell_measure_match(self):
        (cube1,) = iris.load_raw(self.fname)
        cube1 = cube1[:, :, 0, 0]
        cm_and_dims = cube1._cell_measures_and_dims
        (cube2,) = iris.load_raw(self.fname)
        cube2 = cube2[:, :, 0, 0]
        cube2.coord("time").points = cube2.coord("time").points + 1
        cubes = CubeList([cube1, cube2]).concatenate()
        self.assertEqual(cubes[0]._cell_measures_and_dims, cm_and_dims)
        self.assertEqual(len(cubes), 1)

    def test_round_trip(self):
        (cube,) = iris.load(self.fname)
        with self.temp_filename(suffix=".nc") as filename:
            iris.save(cube, filename, unlimited_dimensions=[])
            (round_cube,) = iris.load_raw(filename)
            self.assertEqual(len(round_cube.cell_measures()), 1)
            self.assertEqual(round_cube.cell_measures()[0].measure, "area")

    def test_print(self):
        cube = iris.load_cube(self.fname)
        printed = cube.__str__()
        self.assertTrue(
            (
                "\n     Cell measures:\n          cell_area"
                "                           -         -    "
                "    x         x"
            )
            in printed
        )


@tests.skip_data
class TestCMIP6VolcelloLoad(tests.IrisTest):
    def setUp(self):
        self.fname = tests.get_data_path(
            (
                "NetCDF",
                "volcello",
                "volcello_Ofx_CESM2_deforest-globe_r1i1p1f1_gn.nc",
            )
        )

    def test_cmip6_volcello_load_issue_3367(self):
        # Ensure that reading a file which references itself in
        # `cell_measures` can be read. At the same time, ensure that we
        # still receive a warning about other variables mentioned in
        # `cell_measures` i.e. a warning should be raised about missing
        # areacello.
        areacello_str = "areacello"
        volcello_str = "volcello"
        expected_msg = (
            "Missing CF-netCDF measure variable %r, "
            "referenced by netCDF variable %r" % (areacello_str, volcello_str)
        )

        with mock.patch("warnings.warn") as warn:
            # ensure file loads without failure
            cube = iris.load_cube(self.fname)
            warn.assert_has_calls([mock.call(expected_msg)])

        # extra check to ensure correct variable was found
        assert cube.standard_name == "ocean_volume"


class TestSelfReferencingVarLoad(tests.IrisTest):
    def setUp(self):
        self.temp_dir_path = os.path.join(
            tempfile.mkdtemp(), "issue_3367_volcello_test_file.nc"
        )
        dataset = nc.Dataset(self.temp_dir_path, "w")

        dataset.createDimension("lat", 4)
        dataset.createDimension("lon", 5)
        dataset.createDimension("lev", 3)

        latitudes = dataset.createVariable("lat", np.float64, ("lat",))
        longitudes = dataset.createVariable("lon", np.float64, ("lon",))
        levels = dataset.createVariable("lev", np.float64, ("lev",))
        volcello = dataset.createVariable(
            "volcello", np.float32, ("lat", "lon", "lev")
        )

        latitudes.standard_name = "latitude"
        latitudes.units = "degrees_north"
        latitudes.axis = "Y"
        latitudes[:] = np.linspace(-90, 90, 4)

        longitudes.standard_name = "longitude"
        longitudes.units = "degrees_east"
        longitudes.axis = "X"
        longitudes[:] = np.linspace(0, 360, 5)

        levels.standard_name = "olevel"
        levels.units = "centimeters"
        levels.positive = "down"
        levels.axis = "Z"
        levels[:] = np.linspace(0, 10 ** 5, 3)

        volcello.id = "volcello"
        volcello.out_name = "volcello"
        volcello.standard_name = "ocean_volume"
        volcello.units = "m3"
        volcello.realm = "ocean"
        volcello.frequency = "fx"
        volcello.cell_measures = "area: areacello volume: volcello"
        volcello = np.arange(4 * 5 * 3).reshape((4, 5, 3))

        dataset.close()

    def test_self_referencing_load_issue_3367(self):
        # Ensure that reading a file which references itself in
        # `cell_measures` can be read. At the same time, ensure that we
        # still receive a warning about other variables mentioned in
        # `cell_measures` i.e. a warning should be raised about missing
        # areacello.
        areacello_str = "areacello"
        volcello_str = "volcello"
        expected_msg = (
            "Missing CF-netCDF measure variable %r, "
            "referenced by netCDF variable %r" % (areacello_str, volcello_str)
        )

        with mock.patch("warnings.warn") as warn:
            # ensure file loads without failure
            cube = iris.load_cube(self.temp_dir_path)
            warn.assert_called_with(expected_msg)

        # extra check to ensure correct variable was found
        assert cube.standard_name == "ocean_volume"

    def tearDown(self):
        os.remove(self.temp_dir_path)


class TestCellMethod_unknown(tests.IrisTest):
    def test_unknown_method(self):
        cube = Cube([1, 2], long_name="odd_phenomenon")
        cube.add_cell_method(CellMethod(method="oddity", coords=("x",)))
        temp_dirpath = tempfile.mkdtemp()
        try:
            temp_filepath = os.path.join(temp_dirpath, "tmp.nc")
            iris.save(cube, temp_filepath)
            with warnings.catch_warnings(record=True) as warning_records:
                iris.load(temp_filepath)
            # Filter to get the warning we are interested in.
            warning_messages = [record.message for record in warning_records]
            warning_messages = [
                warn
                for warn in warning_messages
                if isinstance(warn, UnknownCellMethodWarning)
            ]
            self.assertEqual(len(warning_messages), 1)
            message = warning_messages[0].args[0]
            msg = (
                "NetCDF variable 'odd_phenomenon' contains unknown cell "
                "method 'oddity'"
            )
            self.assertIn(msg, message)
        finally:
            shutil.rmtree(temp_dirpath)


@tests.skip_data
class TestCoordSystem(tests.IrisTest):
    def test_load_laea_grid(self):
        cube = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "lambert_azimuthal_equal_area", "euro_air_temp.nc")
            )
        )
        self.assertCML(cube, ("netcdf", "netcdf_laea.cml"))


def _get_scale_factor_add_offset(cube, datatype):
    """Utility function used by netCDF data packing tests."""
    if isinstance(datatype, dict):
        dt = np.dtype(datatype["dtype"])
    else:
        dt = np.dtype(datatype)
    cmax = cube.data.max()
    cmin = cube.data.min()
    n = dt.itemsize * 8
    if ma.isMaskedArray(cube.data):
        masked = True
    else:
        masked = False
    if masked:
        scale_factor = (cmax - cmin) / (2 ** n - 2)
    else:
        scale_factor = (cmax - cmin) / (2 ** n - 1)
    if dt.kind == "u":
        add_offset = cmin
    elif dt.kind == "i":
        if masked:
            add_offset = (cmax + cmin) / 2
        else:
            add_offset = cmin + 2 ** (n - 1) * scale_factor
    return (scale_factor, add_offset)


@tests.skip_data
class TestPackedData(tests.IrisTest):
    def _single_test(self, datatype, CDLfilename, manual=False):
        # Read PP input file.
        file_in = tests.get_data_path(
            (
                "PP",
                "cf_processing",
                "000003000000.03.236.000128.1990.12.01.00.00.b.pp",
            )
        )
        cube = iris.load_cube(file_in)
        scale_factor, offset = _get_scale_factor_add_offset(cube, datatype)
        if manual:
            packspec = dict(
                dtype=datatype, scale_factor=scale_factor, add_offset=offset
            )
        else:
            packspec = datatype
        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cube, file_out, packing=packspec)
            decimal = int(-np.log10(scale_factor))
            packedcube = iris.load_cube(file_out)
            # Check that packed cube is accurate to expected precision
            self.assertArrayAlmostEqual(
                cube.data, packedcube.data, decimal=decimal
            )
            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out,
                ("integration", "netcdf", "TestPackedData", CDLfilename),
            )

    def test_single_packed_signed(self):
        """Test saving a single CF-netCDF file with packing."""
        self._single_test("i2", "single_packed_signed.cdl")

    def test_single_packed_unsigned(self):
        """Test saving a single CF-netCDF file with packing into unsigned."""
        self._single_test("u1", "single_packed_unsigned.cdl")

    def test_single_packed_manual_scale(self):
        """Test saving a single CF-netCDF file with packing with scale
        factor and add_offset set manually."""
        self._single_test("i2", "single_packed_manual.cdl", manual=True)

    def _multi_test(self, CDLfilename, multi_dtype=False):
        """Test saving multiple packed cubes with pack_dtype list."""
        # Read PP input file.
        file_in = tests.get_data_path(
            ("PP", "cf_processing", "abcza_pa19591997_daily_29.b.pp")
        )
        cubes = iris.load(file_in)
        # ensure cube order is the same:
        cubes.sort(key=lambda cube: cube.cell_methods[0].method)
        datatype = "i2"
        scale_factor, offset = _get_scale_factor_add_offset(cubes[0], datatype)
        if multi_dtype:
            packdict = dict(
                dtype=datatype, scale_factor=scale_factor, add_offset=offset
            )
            packspec = [packdict, None, "u2"]
            dtypes = packspec
        else:
            packspec = datatype
            dtypes = repeat(packspec)

        # Write Cube to netCDF file.
        with self.temp_filename(suffix=".nc") as file_out:
            iris.save(cubes, file_out, packing=packspec)
            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_out,
                ("integration", "netcdf", "TestPackedData", CDLfilename),
            )
            packedcubes = iris.load(file_out)
            packedcubes.sort(key=lambda cube: cube.cell_methods[0].method)
            for cube, packedcube, dtype in zip(cubes, packedcubes, dtypes):
                if dtype:
                    sf, ao = _get_scale_factor_add_offset(cube, dtype)
                    decimal = int(-np.log10(sf))
                    # Check that packed cube is accurate to expected precision
                    self.assertArrayAlmostEqual(
                        cube.data, packedcube.data, decimal=decimal
                    )
                else:
                    self.assertArrayEqual(cube.data, packedcube.data)

    def test_multi_packed_single_dtype(self):
        """Test saving multiple packed cubes with the same pack_dtype."""
        # Read PP input file.
        self._multi_test("multi_packed_single_dtype.cdl")

    def test_multi_packed_multi_dtype(self):
        """Test saving multiple packed cubes with pack_dtype list."""
        # Read PP input file.
        self._multi_test("multi_packed_multi_dtype.cdl", multi_dtype=True)


class TestScalarCube(tests.IrisTest):
    def test_scalar_cube_save_load(self):
        cube = iris.cube.Cube(1, long_name="scalar_cube")
        with self.temp_filename(suffix=".nc") as fout:
            iris.save(cube, fout)
            scalar_cube = iris.load_cube(fout)
            self.assertEqual(scalar_cube.name(), "scalar_cube")


class TestStandardName(tests.IrisTest):
    def test_standard_name_roundtrip(self):
        standard_name = "air_temperature detection_minimum"
        cube = iris.cube.Cube(1, standard_name=standard_name)
        with self.temp_filename(suffix=".nc") as fout:
            iris.save(cube, fout)
            detection_limit_cube = iris.load_cube(fout)
            self.assertEqual(detection_limit_cube.standard_name, standard_name)


class TestLoadMinimalGeostationary(tests.IrisTest):
    """
    Check we can load data with a geostationary grid-mapping, even when the
    'false-easting' and 'false_northing' properties are missing.

    """

    _geostationary_problem_cdl = """
netcdf geostationary_problem_case {
dimensions:
    y = 2 ;
    x = 3 ;
variables:
    short radiance(y, x) ;
        radiance:standard_name = "toa_outgoing_radiance_per_unit_wavelength" ;
        radiance:units = "W m-2 sr-1 um-1" ;
        radiance:coordinates = "y x" ;
        radiance:grid_mapping = "imager_grid_mapping" ;
    short y(y) ;
        y:units = "rad" ;
        y:axis = "Y" ;
        y:long_name = "fixed grid projection y-coordinate" ;
        y:standard_name = "projection_y_coordinate" ;
    short x(x) ;
        x:units = "rad" ;
        x:axis = "X" ;
        x:long_name = "fixed grid projection x-coordinate" ;
        x:standard_name = "projection_x_coordinate" ;
    int imager_grid_mapping ;
        imager_grid_mapping:grid_mapping_name = "geostationary" ;
        imager_grid_mapping:perspective_point_height = 35786023. ;
        imager_grid_mapping:semi_major_axis = 6378137. ;
        imager_grid_mapping:semi_minor_axis = 6356752.31414 ;
        imager_grid_mapping:latitude_of_projection_origin = 0. ;
        imager_grid_mapping:longitude_of_projection_origin = -75. ;
        imager_grid_mapping:sweep_angle_axis = "x" ;

data:

 // coord values, just so these can be dim-coords
 y = 0, 1 ;
 x = 0, 1, 2 ;

}
"""

    @classmethod
    def setUpClass(cls):
        # Create a temp directory for transient test files.
        cls.temp_dir = tempfile.mkdtemp()
        cls.path_test_cdl = path_join(cls.temp_dir, "geos_problem.cdl")
        cls.path_test_nc = path_join(cls.temp_dir, "geos_problem.nc")
        # Create a reference file from the CDL text.
        with open(cls.path_test_cdl, "w") as f_out:
            f_out.write(cls._geostationary_problem_cdl)
        # Call 'ncgen' to make an actual netCDF file from the CDL.
        command = "ncgen -o {} {}".format(cls.path_test_nc, cls.path_test_cdl)
        check_call(command, shell=True)

    @classmethod
    def tearDownClass(cls):
        # Destroy the temp directory.
        shutil.rmtree(cls.temp_dir)

    def test_geostationary_no_false_offsets(self):
        # Check we can load the test data and coordinate system properties are correct.
        cube = iris.load_cube(self.path_test_nc)
        # Check the coordinate system properties has the correct default properties.
        cs = cube.coord_system()
        self.assertIsInstance(cs, iris.coord_systems.Geostationary)
        self.assertEqual(cs.false_easting, 0.0)
        self.assertEqual(cs.false_northing, 0.0)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests
import iris.coords

import os
import tempfile

import netCDF4

import iris
import iris.tests.pp as pp
import iris.util
from iris.fileformats.pp import STASH


def callback_000003000000_16_202_000128_1860_09_01_00_00_b_pp(
    cube, field, filename
):
    cube.attributes["STASH"] = STASH(1, 16, 202)
    cube.standard_name = "geopotential_height"
    cube.units = "m"


def callback_HadCM2_ts_SAT_ann_18602100_b_pp(cube, field, filename):
    def reset_pole(coord_name):
        coord = cube.coord(coord_name)
        coord.rename(coord.name().replace("grid_", ""))
        coord.coord_system = coord.coord_system.ellipsoid

    reset_pole("grid_latitude")
    reset_pole("grid_longitude")
    cube.standard_name = "air_temperature"
    cube.units = "Celsius"
    cube.attributes["STASH"] = STASH(1, 3, 236)
    # Force the height to 1.5m
    if cube.coords("height"):
        cube.remove_coord("height")
    height_coord = iris.coords.DimCoord(1.5, standard_name="height", units="m")
    cube.add_aux_coord(height_coord)


def callback_model_b_pp(cube, field, filename):
    cube.standard_name = "air_temperature"
    cube.units = "K"
    cube.attributes["STASH"] = STASH(1, 16, 203)


def callback_integer_b_pp(cube, field, filename):
    cube.standard_name = "land_binary_mask"
    cube.units = "1"
    del cube.attributes["STASH"]


def callback_001000000000_00_000_000000_1860_01_01_00_00_f_b_pp(
    cube, field, filename
):
    cube.standard_name = "sea_surface_height_above_geoid"
    cube.units = "m"


def callback_aaxzc_n10r13xy_b_pp(cube, field, filename):
    height_coord = iris.coords.DimCoord(1.5, long_name="height", units="m")
    cube.add_aux_coord(height_coord)


@tests.skip_data
class TestAll(tests.IrisTest, pp.PPTest):
    _ref_dir = ("usecases", "pp_to_cf_conversion")

    def _test_file(self, name):
        """This is the main test routine that is called for each of the files listed below."""
        pp_path = self._src_pp_path(name)

        # 1) Load the PP and check the Cube
        callback_name = "callback_" + name.replace(".", "_")
        callback = globals().get(callback_name)
        cubes = iris.load(pp_path, callback=callback)

        if name.endswith(".pp"):
            fname_name = name[:-3]
        else:
            fname_name = name

        self.assertCML(cubes, self._ref_dir + ("from_pp", fname_name + ".cml"))

        # 2) Save the Cube and check the netCDF
        nc_filenames = []

        for index, cube in enumerate(cubes):
            # Explicitly set a fill-value as a workaround for
            # https://github.com/Unidata/netcdf4-python/issues/725
            fill_value = netCDF4.default_fillvals[cube.dtype.str[1:]]

            file_nc = tempfile.NamedTemporaryFile(
                suffix=".nc", delete=False
            ).name
            iris.save(
                cube,
                file_nc,
                netcdf_format="NETCDF3_CLASSIC",
                fill_value=fill_value,
            )

            # Check the netCDF file against CDL expected output.
            self.assertCDL(
                file_nc,
                self._ref_dir
                + ("to_netcdf", "%s_%d.cdl" % (fname_name, index)),
            )
            nc_filenames.append(file_nc)

        # 3) Load the netCDF and check the Cube
        for index, nc_filename in enumerate(nc_filenames):
            # Read netCDF to Cube.
            cube = iris.load_cube(nc_filename)
            self.assertCML(
                cube,
                self._ref_dir
                + ("from_netcdf", "%s_%d.cml" % (fname_name, index)),
            )
            os.remove(nc_filename)

        # 4) Save the Cube and check the PP
        # Only the first four files pass their tests at the moment.

        if name in self.files_to_check[:4]:
            self._test_pp_save(cubes, name)

    def _src_pp_path(self, name):
        return tests.get_data_path(("PP", "cf_processing", name))

    def _test_pp_save(self, cubes, name):
        # If there's no existing reference file then make it from the *source* data
        reference_txt_path = tests.get_result_path(
            self._ref_dir + ("to_pp", name + ".txt")
        )
        reference_pp_path = self._src_pp_path(name)
        with self.cube_save_test(
            reference_txt_path, reference_pp_path=reference_pp_path
        ) as temp_pp_path:
            iris.save(cubes, temp_pp_path)

    files_to_check = [
        "000003000000.03.236.000128.1990.12.01.00.00.b.pp",
        "000003000000.03.236.004224.1990.12.01.00.00.b.pp",
        "000003000000.03.236.008320.1990.12.01.00.00.b.pp",
        "000003000000.16.202.000128.1860.09.01.00.00.b.pp",
        "001000000000.00.000.000000.1860.01.01.00.00.f.b.pp",
        "002000000000.44.101.131200.1920.09.01.00.00.b.pp",
        "008000000000.44.101.000128.1890.09.01.00.00.b.pp",
        "HadCM2_ts_SAT_ann_18602100.b.pp",
        "aaxzc_level_lat_orig.b.pp",
        "aaxzc_lon_lat_press_orig.b.pp",
        "abcza_pa19591997_daily_29.b.pp",
        "12187.b.pp",
        "ocean_xsect.b.pp",
        "model.b.pp",
        "integer.b.pp",
        "aaxzc_lon_lat_several.b.pp",
        "aaxzc_n10r13xy.b.pp",
        "aaxzc_time_press.b.pp",
        "aaxzc_tseries.b.pp",
        "abxpa_press_lat.b.pp",
        "st30211.b.pp",
        "st0fc942.b.pp",
        "st0fc699.b.pp",
    ]


def make_test_function(func_name, file_name):
    """Builds a function which can be later turned into a bound method."""
    scope = {}
    exec(
        """def %s(self):
                name = %r
                self._test_file(name)
    """
        % (func_name, file_name),
        scope,
        scope,
    )
    # return the newly created function
    return scope[func_name]


def attach_tests():
    # attach a test method on TestAll for each file to test
    for file_name in TestAll.files_to_check:
        func_name = "test_{}".format(file_name.replace(".", "_"))
        test_func = make_test_function(func_name, file_name)
        setattr(TestAll, func_name, test_func)


attach_tests()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.common.metadata.metadata_manager_factory`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import pickle
import unittest.mock as mock

from cf_units import Unit

from iris.common.metadata import (
    AncillaryVariableMetadata,
    BaseMetadata,
    CellMeasureMetadata,
    CoordMetadata,
    CubeMetadata,
    metadata_manager_factory,
)


BASES = [
    AncillaryVariableMetadata,
    BaseMetadata,
    CellMeasureMetadata,
    CoordMetadata,
    CubeMetadata,
]


class Test_factory(tests.IrisTest):
    def test__subclass_invalid(self):
        class Other:
            pass

        emsg = "Require a subclass of 'BaseMetadata'"
        with self.assertRaisesRegex(TypeError, emsg):
            _ = metadata_manager_factory(Other)

    def test__kwargs_invalid(self):
        emsg = "Invalid 'BaseMetadata' field parameters, got 'wibble'."
        with self.assertRaisesRegex(ValueError, emsg):
            metadata_manager_factory(BaseMetadata, wibble="nope")


class Test_instance(tests.IrisTest):
    def setUp(self):
        self.bases = BASES

    def test__namespace(self):
        namespace = [
            "DEFAULT_NAME",
            "__init__",
            "__eq__",
            "__getstate__",
            "__ne__",
            "__reduce__",
            "__repr__",
            "__setstate__",
            "fields",
            "name",
            "token",
            "values",
        ]
        for base in self.bases:
            metadata = metadata_manager_factory(base)
            for name in namespace:
                self.assertTrue(hasattr(metadata, name))
            if base is CubeMetadata:
                self.assertTrue(hasattr(metadata, "_names"))
            self.assertIs(metadata.cls, base)

    def test__kwargs_default(self):
        for base in self.bases:
            kwargs = dict(zip(base._fields, [None] * len(base._fields)))
            metadata = metadata_manager_factory(base)
            self.assertEqual(metadata.values._asdict(), kwargs)

    def test__kwargs(self):
        for base in self.bases:
            kwargs = dict(zip(base._fields, range(len(base._fields))))
            metadata = metadata_manager_factory(base, **kwargs)
            self.assertEqual(metadata.values._asdict(), kwargs)


class Test_instance___eq__(tests.IrisTest):
    def setUp(self):
        self.metadata = metadata_manager_factory(BaseMetadata)

    def test__not_implemented(self):
        self.assertNotEqual(self.metadata, 1)

    def test__not_is_cls(self):
        base = BaseMetadata
        other = metadata_manager_factory(base)
        self.assertIs(other.cls, base)
        other.cls = CoordMetadata
        self.assertNotEqual(self.metadata, other)

    def test__not_values(self):
        standard_name = mock.sentinel.standard_name
        other = metadata_manager_factory(
            BaseMetadata, standard_name=standard_name
        )
        self.assertEqual(other.standard_name, standard_name)
        self.assertIsNone(other.long_name)
        self.assertIsNone(other.var_name)
        self.assertIsNone(other.units)
        self.assertIsNone(other.attributes)
        self.assertNotEqual(self.metadata, other)

    def test__same_default(self):
        other = metadata_manager_factory(BaseMetadata)
        self.assertEqual(self.metadata, other)

    def test__same(self):
        kwargs = dict(
            standard_name=1, long_name=2, var_name=3, units=4, attributes=5
        )
        metadata = metadata_manager_factory(BaseMetadata, **kwargs)
        other = metadata_manager_factory(BaseMetadata, **kwargs)
        self.assertEqual(metadata.values._asdict(), kwargs)
        self.assertEqual(metadata, other)


class Test_instance____repr__(tests.IrisTest):
    def setUp(self):
        self.metadata = metadata_manager_factory(BaseMetadata)

    def test(self):
        standard_name = mock.sentinel.standard_name
        long_name = mock.sentinel.long_name
        var_name = mock.sentinel.var_name
        units = mock.sentinel.units
        attributes = mock.sentinel.attributes
        values = (standard_name, long_name, var_name, units, attributes)

        for field, value in zip(self.metadata.fields, values):
            setattr(self.metadata, field, value)

        result = repr(self.metadata)
        expected = (
            "MetadataManager(standard_name={!r}, long_name={!r}, var_name={!r}, "
            "units={!r}, attributes={!r})"
        )
        self.assertEqual(result, expected.format(*values))


class Test_instance__pickle(tests.IrisTest):
    def setUp(self):
        self.standard_name = "standard_name"
        self.long_name = "long_name"
        self.var_name = "var_name"
        self.units = Unit("1")
        self.attributes = dict(hello="world")
        values = (
            self.standard_name,
            self.long_name,
            self.var_name,
            self.units,
            self.attributes,
        )
        self.kwargs = dict(zip(BaseMetadata._fields, values))
        self.metadata = metadata_manager_factory(BaseMetadata, **self.kwargs)

    def test_pickle(self):
        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            with self.temp_filename(suffix=".pkl") as fname:
                with open(fname, "wb") as fo:
                    pickle.dump(self.metadata, fo, protocol=protocol)
                with open(fname, "rb") as fi:
                    metadata = pickle.load(fi)
                    self.assertEqual(metadata, self.metadata)


class Test_instance__fields(tests.IrisTest):
    def setUp(self):
        self.bases = BASES

    def test(self):
        for base in self.bases:
            fields = base._fields
            metadata = metadata_manager_factory(base)
            self.assertEqual(metadata.fields, fields)
            for field in fields:
                hasattr(metadata, field)


class Test_instance__values(tests.IrisTest):
    def setUp(self):
        self.bases = BASES

    def test(self):
        for base in self.bases:
            metadata = metadata_manager_factory(base)
            result = metadata.values
            self.assertIsInstance(result, base)
            self.assertEqual(result._fields, base._fields)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.analysis.PointInCell`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.analysis import PointInCell


class Test_regridder(tests.IrisTest):
    def test(self):
        point_in_cell = PointInCell(mock.sentinel.weights)

        with mock.patch(
            "iris.analysis.CurvilinearRegridder",
            return_value=mock.sentinel.regridder,
        ) as ecr:
            regridder = point_in_cell.regridder(
                mock.sentinel.src, mock.sentinel.target
            )

        ecr.assert_called_once_with(
            mock.sentinel.src, mock.sentinel.target, mock.sentinel.weights
        )
        self.assertIs(regridder, mock.sentinel.regridder)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_auxilliary_coordinate`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.coords import AuxCoord
from iris.fileformats.cf import CFVariable
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_auxiliary_coordinate


# from iris.tests.unit.fileformats.pyke_rules.compiled_krb\
#     .fc_rules_cf_fc.test_build_dimension_coordinate import RulesTestMixin

class TestBoundsVertexDim(tests.IrisTest):
    # Lookup for various tests (which change the dimension order).
    dim_names_lens = {
        'foo': 2, 'bar': 3, 'nv': 4,
        # 'x' and 'y' used as aliases for 'foo' and 'bar'
        'x': 2, 'y': 3}

    def setUp(self):
        # Create coordinate cf variables and pyke engine.
        dimension_names = ('foo', 'bar')
        points, cf_data = self._make_array_and_cf_data(dimension_names)
        self.cf_coord_var = mock.Mock(
            spec=CFVariable,
            dimensions=dimension_names,
            cf_name='wibble',
            cf_data=cf_data,
            standard_name=None,
            long_name='wibble',
            units='m',
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

        expected_bounds, _ = self._make_array_and_cf_data(
            dimension_names=('foo', 'bar', 'nv'))
        self.expected_coord = AuxCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=expected_bounds)

        self.engine = mock.Mock(
            cube=mock.Mock(),
            cf_var=mock.Mock(dimensions=('foo', 'bar'),
                             cf_data=cf_data),
            filename='DUMMY',
            cube_parts=dict(coordinates=[]))

        # Patch the deferred loading that prevents attempted file access.
        # This assumes that self.cf_bounds_var is defined in the test case.
        def patched__getitem__(proxy_self, keys):
            for var in (self.cf_coord_var, self.cf_bounds_var):
                if proxy_self.variable_name == var.cf_name:
                    return var[keys]
            raise RuntimeError()

        self.patch('iris.fileformats.netcdf.NetCDFDataProxy.__getitem__',
                   new=patched__getitem__)

        # Patch the helper function that retrieves the bounds cf variable,
        # and a False flag for climatological.
        # This avoids the need for setting up further mocking of cf objects.
        def _get_per_test_bounds_var(_coord_unused):
            # Return the 'cf_bounds_var' created by the current test.
            return (self.cf_bounds_var, False)

        self.patch('iris.fileformats._pyke_rules.compiled_krb.'
                   'fc_rules_cf_fc.get_cf_bounds_var',
                   new=_get_per_test_bounds_var)

    @classmethod
    def _make_array_and_cf_data(cls, dimension_names):
        shape = tuple(cls.dim_names_lens[name]
                      for name in dimension_names)
        cf_data = mock.MagicMock(_FillValue=None, spec=[])
        cf_data.chunking = mock.MagicMock(return_value=shape)
        return np.zeros(shape), cf_data

    def _make_cf_bounds_var(self, dimension_names):
        # Create the bounds cf variable.
        bounds, cf_data = self._make_array_and_cf_data(dimension_names)
        cf_bounds_var = mock.Mock(
            spec=CFVariable,
            dimensions=dimension_names,
            cf_name='wibble_bnds',
            cf_data=cf_data,
            shape=bounds.shape,
            dtype=bounds.dtype,
            __getitem__=lambda self, key: bounds[key])

        return bounds, cf_bounds_var

    def _check_case(self, dimension_names):
        bounds, self.cf_bounds_var = self._make_cf_bounds_var(
            dimension_names=dimension_names)

        # Asserts must lie within context manager because of deferred loading.
        build_auxiliary_coordinate(self.engine, self.cf_coord_var)

        # Test that expected coord is built and added to cube.
        self.engine.cube.add_aux_coord.assert_called_with(
            self.expected_coord, [0, 1])

        # Test that engine.cube_parts container is correctly populated.
        expected_list = [(self.expected_coord, self.cf_coord_var.cf_name)]
        self.assertEqual(self.engine.cube_parts['coordinates'],
                         expected_list)

    def test_fastest_varying_vertex_dim(self):
        # The usual order.
        self._check_case(dimension_names=('foo', 'bar', 'nv'))

    def test_slowest_varying_vertex_dim(self):
        # Bounds in the first (slowest varying) dimension.
        self._check_case(dimension_names=('nv', 'foo', 'bar'))

    def test_fastest_with_different_dim_names(self):
        # Despite the dimension names ('x', and 'y') differing from the coord's
        # which are 'foo' and 'bar' (as permitted by the cf spec),
        # this should still work because the vertex dim is the fastest varying.
        self._check_case(dimension_names=('x', 'y', 'nv'))


class TestDtype(tests.IrisTest):
    def setUp(self):
        # Create coordinate cf variables and pyke engine.
        points = np.arange(6).reshape(2, 3)
        cf_data = mock.MagicMock(_FillValue=None)
        cf_data.chunking = mock.MagicMock(return_value=points.shape)

        self.cf_coord_var = mock.Mock(
            spec=CFVariable,
            dimensions=('foo', 'bar'),
            cf_name='wibble',
            cf_data=cf_data,
            standard_name=None,
            long_name='wibble',
            units='m',
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

        self.engine = mock.Mock(
            cube=mock.Mock(),
            cf_var=mock.Mock(dimensions=('foo', 'bar')),
            filename='DUMMY',
            cube_parts=dict(coordinates=[]))

        def patched__getitem__(proxy_self, keys):
            if proxy_self.variable_name == self.cf_coord_var.cf_name:
                return self.cf_coord_var[keys]
            raise RuntimeError()

        self.deferred_load_patch = mock.patch(
            'iris.fileformats.netcdf.NetCDFDataProxy.__getitem__',
            new=patched__getitem__)

    def test_scale_factor_add_offset_int(self):
        self.cf_coord_var.scale_factor = 3
        self.cf_coord_var.add_offset = 5

        with self.deferred_load_patch:
            build_auxiliary_coordinate(self.engine, self.cf_coord_var)

        coord, _ = self.engine.cube_parts['coordinates'][0]
        self.assertEqual(coord.dtype.kind, 'i')

    def test_scale_factor_float(self):
        self.cf_coord_var.scale_factor = 3.

        with self.deferred_load_patch:
            build_auxiliary_coordinate(self.engine, self.cf_coord_var)

        coord, _ = self.engine.cube_parts['coordinates'][0]
        self.assertEqual(coord.dtype.kind, 'f')

    def test_add_offset_float(self):
        self.cf_coord_var.add_offset = 5.

        with self.deferred_load_patch:
            build_auxiliary_coordinate(self.engine, self.cf_coord_var)

        coord, _ = self.engine.cube_parts['coordinates'][0]
        self.assertEqual(coord.dtype.kind, 'f')


class TestCoordConstruction(tests.IrisTest):
    def setUp(self):
        # Create dummy pyke engine.
        self.engine = mock.Mock(
            cube=mock.Mock(),
            cf_var=mock.Mock(dimensions=('foo', 'bar')),
            filename='DUMMY',
            cube_parts=dict(coordinates=[]))

        points = np.arange(6)
        self.cf_coord_var = mock.Mock(
            dimensions=('foo',),
            scale_factor=1,
            add_offset=0,
            cf_name='wibble',
            cf_data=mock.MagicMock(chunking=mock.Mock(return_value=None), spec=[]),
            standard_name=None,
            long_name='wibble',
            units='days since 1970-01-01',
            calendar=None,
            shape=points.shape,
            dtype=points.dtype,
            __getitem__=lambda self, key: points[key])

        bounds = np.arange(12).reshape(6, 2)
        self.cf_bounds_var = mock.Mock(
            dimensions=('x', 'nv'),
            scale_factor=1,
            add_offset=0,
            cf_name='wibble_bnds',
            cf_data=mock.MagicMock(chunking=mock.Mock(return_value=None)),
            shape=bounds.shape,
            dtype=bounds.dtype,
            __getitem__=lambda self, key: bounds[key])
        self.bounds = bounds

        # Create patch for deferred loading that prevents attempted
        # file access. This assumes that self.cf_coord_var and
        # self.cf_bounds_var are defined in the test case.
        def patched__getitem__(proxy_self, keys):
            for var in (self.cf_coord_var, self.cf_bounds_var):
                if proxy_self.variable_name == var.cf_name:
                    return var[keys]
            raise RuntimeError()

        self.patch('iris.fileformats.netcdf.NetCDFDataProxy.__getitem__',
                   new=patched__getitem__)

        # Patch the helper function that retrieves the bounds cf variable.
        # This avoids the need for setting up further mocking of cf objects.
        self.use_climatology_bounds = False  # Set this when you need to.

        def get_cf_bounds_var(coord_var):
            return self.cf_bounds_var, self.use_climatology_bounds

        self.patch('iris.fileformats._pyke_rules.compiled_krb.'
                   'fc_rules_cf_fc.get_cf_bounds_var',
                   new=get_cf_bounds_var)

    def check_case_aux_coord_construction(self, climatology=False):
        # Test a generic auxiliary coordinate, with or without
        # a climatological coord.
        self.use_climatology_bounds = climatology

        expected_coord = AuxCoord(
            self.cf_coord_var[:],
            long_name=self.cf_coord_var.long_name,
            var_name=self.cf_coord_var.cf_name,
            units=self.cf_coord_var.units,
            bounds=self.bounds,
            climatological=climatology)

        build_auxiliary_coordinate(self.engine, self.cf_coord_var)

        # Test that expected coord is built and added to cube.
        self.engine.cube.add_aux_coord.assert_called_with(
            expected_coord, [0])

    def test_aux_coord_construction(self):
        self.check_case_aux_coord_construction(climatology=False)

    def test_aux_coord_construction__climatology(self):
        self.check_case_aux_coord_construction(climatology=True)


if __name__ == '__main__':
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

import iris
import iris.fileformats.nimrod_load_rules as nimrod_load_rules
from iris.exceptions import TranslationError


def mock_nimrod_field():
    field = iris.fileformats.nimrod.NimrodField()
    field.int_mdi = -32767
    field.float32_mdi = -32767.0
    return field


class TestLoad(tests.IrisTest):
    @tests.skip_data
    def test_multi_field_load(self):
        # load a cube with two fields
        cube = iris.load(
            tests.get_data_path(
                (
                    "NIMROD",
                    "uk2km",
                    "WO0000000003452",
                    "201007020900_u1096_ng_ey00_visibility0180_screen_2km",
                )
            )
        )
        self.assertCML(cube, ("nimrod", "load_2flds.cml"))

    @tests.skip_data
    def test_huge_field_load(self):
        # load a wide range of cubes with all meta-data variations
        for datafile in {
            "u1096_ng_ek07_precip0540_accum180_18km",
            "u1096_ng_ek00_cloud3d0060_2km",
            "u1096_ng_ek00_cloud_2km",
            "u1096_ng_ek00_convection_2km",
            "u1096_ng_ek00_convwind_2km",
            "u1096_ng_ek00_frzlev_2km",
            "u1096_ng_ek00_height_2km",
            "u1096_ng_ek00_precip_2km",
            "u1096_ng_ek00_precipaccum_2km",
            "u1096_ng_ek00_preciptype_2km",
            "u1096_ng_ek00_pressure_2km",
            "u1096_ng_ek00_radiation_2km",
            "u1096_ng_ek00_radiationuv_2km",
            "u1096_ng_ek00_refl_2km",
            "u1096_ng_ek00_relhumidity3d0060_2km",
            "u1096_ng_ek00_relhumidity_2km",
            "u1096_ng_ek00_snow_2km",
            "u1096_ng_ek00_soil3d0060_2km",
            "u1096_ng_ek00_soil_2km",
            "u1096_ng_ek00_temperature_2km",
            "u1096_ng_ek00_visibility_2km",
            "u1096_ng_ek00_wind_2km",
            "u1096_ng_ek00_winduv3d0015_2km",
            "u1096_ng_ek00_winduv_2km",
            "u1096_ng_ek01_cape_2km",
            "u1096_ng_umqv_fog_2km",
            "u1096_ng_bmr04_precip_2km",
            "u1096_ng_bsr05_precip_accum60_2km",
            "probability_fields",
        }:
            cube = iris.load(
                tests.get_data_path(("NIMROD", "uk2km", "cutouts", datafile))
            )
            self.assertCML(cube, ("nimrod", f"{datafile}.cml"))

    @tests.skip_data
    def test_load_kwarg(self):
        """Tests that the handle_metadata_errors kwarg is effective by setting it to
        False with a file with known incomplete meta-data (missing ellipsoid)."""
        datafile = "u1096_ng_ek00_pressure_2km"
        with self.assertRaisesRegex(
            TranslationError,
            "Ellipsoid not supported, proj_biaxial_ellipsoid:-32767, horizontal_grid_type:0",
        ):
            with open(
                tests.get_data_path(("NIMROD", "uk2km", "cutouts", datafile)),
                "rb",
            ) as infile:
                iris.fileformats.nimrod_load_rules.run(
                    iris.fileformats.nimrod.NimrodField(infile),
                    handle_metadata_errors=False,
                )

    def test_orography(self):
        # Mock an orography field we've seen.
        field = mock_nimrod_field()
        cube = iris.cube.Cube(np.arange(100).reshape(10, 10))

        field.dt_year = field.dt_month = field.dt_day = field.int_mdi
        field.dt_hour = field.dt_minute = field.int_mdi
        field.proj_biaxial_ellipsoid = 0
        field.tm_meridian_scaling = 0.999601
        field.field_code = 73
        field.reference_vertical_coord_type = field.int_mdi  # Not bounded
        field.reference_vertical_coord = field.int_mdi
        field.vertical_coord_type = 1
        field.vertical_coord = 8888
        field.ensemble_member = field.int_mdi
        field.threshold_value = field.int_mdi
        field.title = "(MOCK) 2km mean orography"
        field.units = "metres"
        field.source = "GLOBE DTM"

        nimrod_load_rules.name(cube, field, handle_metadata_errors=True)
        nimrod_load_rules.units(cube, field)
        nimrod_load_rules.reference_time(cube, field)
        nimrod_load_rules.vertical_coord(cube, field)
        nimrod_load_rules.attributes(cube, field)

        self.assertCML(cube, ("nimrod", "mockography.cml"))

    def test_levels_below_ground(self):
        # Mock a soil temperature field we've seen.
        field = mock_nimrod_field()
        cube = iris.cube.Cube(np.arange(100).reshape(10, 10))

        field.field_code = -1  # Not orography
        field.reference_vertical_coord_type = field.int_mdi  # Not bounded
        field.reference_vertical_coord = field.int_mdi
        field.vertical_coord_type = 12
        field.vertical_coord = 42
        nimrod_load_rules.vertical_coord(cube, field)

        self.assertCML(cube, ("nimrod", "levels_below_ground.cml"))

    def test_period_of_interest(self):
        # mock a pressure field
        field = mock_nimrod_field()
        cube = iris.cube.Cube(np.arange(100).reshape(10, 10))

        field.field_code = 0
        field.vt_year = 2013
        field.vt_month = 5
        field.vt_day = 7
        field.vt_hour = 6
        field.vt_minute = 0
        field.vt_second = 0
        field.dt_year = 2013
        field.dt_month = 5
        field.dt_day = 7
        field.dt_hour = 6
        field.dt_minute = 0
        field.dt_second = 0
        field.period_minutes = 60

        nimrod_load_rules.time(cube, field)

        self.assertCML(cube, ("nimrod", "period_of_interest.cml"))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides objects for building up expressions useful for pattern matching.

"""

from collections.abc import Iterable, Mapping
import operator

import numpy as np

import iris.coords
import iris.exceptions


class Constraint:
    """
    Constraints are the mechanism by which cubes can be pattern matched and
    filtered according to specific criteria.

    Once a constraint has been defined, it can be applied to cubes using the
    :meth:`Constraint.extract` method.

    """

    def __init__(self, name=None, cube_func=None, coord_values=None, **kwargs):
        """
        Creates a new instance of a Constraint which can be used for filtering
        cube loading or cube list extraction.

        Args:

        * name:   string or None
            If a string, it is used as the name to match against the
            `~iris.cube.Cube.names` property.
        * cube_func:   callable or None
            If a callable, it must accept a Cube as its first and only argument
            and return either True or False.
        * coord_values:   dict or None
            If a dict, it must map coordinate name to the condition on the
            associated coordinate.
        * `**kwargs`:
            The remaining keyword arguments are converted to coordinate
            constraints. The name of the argument gives the name of a
            coordinate, and the value of the argument is the condition to meet
            on that coordinate::

                Constraint(model_level_number=10)

            Coordinate level constraints can be of several types:

            * **string, int or float** - the value of the coordinate to match.
              e.g. ``model_level_number=10``

            * **list of values** - the possible values that the coordinate may
              have to match. e.g. ``model_level_number=[10, 12]``

            * **callable** - a function which accepts a
              :class:`iris.coords.Cell` instance as its first and only argument
              returning True or False if the value of the Cell is desired.
              e.g. ``model_level_number=lambda cell: 5 < cell < 10``

        The :ref:`user guide <loading_iris_cubes>` covers cube much of
        constraining in detail, however an example which uses all of the
        features of this class is given here for completeness::

            Constraint(name='air_potential_temperature',
                       cube_func=lambda cube: cube.units == 'kelvin',
                       coord_values={'latitude':lambda cell: 0 < cell < 90},
                       model_level_number=[10, 12])
                       & Constraint(ensemble_member=2)

        Constraint filtering is performed at the cell level.
        For further details on how cell comparisons are performed see
        :class:`iris.coords.Cell`.

        """
        if not (name is None or isinstance(name, str)):
            raise TypeError("name must be None or string, got %r" % name)
        if not (cube_func is None or callable(cube_func)):
            raise TypeError(
                "cube_func must be None or callable, got %r" % cube_func
            )
        if not (coord_values is None or isinstance(coord_values, Mapping)):
            raise TypeError(
                "coord_values must be None or a "
                "collections.Mapping, got %r" % coord_values
            )

        coord_values = coord_values or {}
        duplicate_keys = set(coord_values.keys()) & set(kwargs.keys())
        if duplicate_keys:
            raise ValueError(
                "Duplicate coordinate conditions specified for: "
                "%s" % list(duplicate_keys)
            )

        self._name = name
        self._cube_func = cube_func

        self._coord_values = coord_values.copy()
        self._coord_values.update(kwargs)

        self._coord_constraints = []
        for coord_name, coord_thing in self._coord_values.items():
            self._coord_constraints.append(
                _CoordConstraint(coord_name, coord_thing)
            )

    def __repr__(self):
        args = []
        if self._name:
            args.append(("name", self._name))
        if self._cube_func:
            args.append(("cube_func", self._cube_func))
        if self._coord_values:
            args.append(("coord_values", self._coord_values))
        return "Constraint(%s)" % ", ".join("%s=%r" % (k, v) for k, v in args)

    def _coordless_match(self, cube):
        """
        Return whether this constraint matches the given cube when not
        taking coordinates into account.

        """
        match = True
        if self._name:
            # Require to also check against cube.name() for the fallback
            # "unknown" default case, when there is no name metadata available.
            match = self._name in cube._names or self._name == cube.name()
        if match and self._cube_func:
            match = self._cube_func(cube)
        return match

    def extract(self, cube):
        """
        Return the subset of the given cube which matches this constraint,
        else return None.

        """
        resultant_CIM = self._CIM_extract(cube)
        slice_tuple = resultant_CIM.as_slice()
        result = None
        if slice_tuple is not None:
            # Slicing the cube is an expensive operation.
            if all([item == slice(None) for item in slice_tuple]):
                # Don't perform a full slice, just return the cube.
                result = cube
            else:
                # Performing the partial slice.
                result = cube[slice_tuple]
        return result

    def _CIM_extract(self, cube):
        # Returns _ColumnIndexManager

        # Cater for scalar cubes by setting the dimensionality to 1
        # when cube.ndim is 0.
        resultant_CIM = _ColumnIndexManager(cube.ndim or 1)

        if not self._coordless_match(cube):
            resultant_CIM.all_false()
        else:
            for coord_constraint in self._coord_constraints:
                resultant_CIM = resultant_CIM & coord_constraint.extract(cube)

        return resultant_CIM

    def __and__(self, other):
        return ConstraintCombination(self, other, operator.__and__)

    def __rand__(self, other):
        return ConstraintCombination(other, self, operator.__and__)


class ConstraintCombination(Constraint):
    """Represents the binary combination of two Constraint instances."""

    def __init__(self, lhs, rhs, operator):
        """
        A ConstraintCombination instance is created by providing two
        Constraint instances and the appropriate :mod:`operator`.

        """
        try:
            lhs_constraint = as_constraint(lhs)
            rhs_constraint = as_constraint(rhs)
        except TypeError:
            raise TypeError(
                "Can only combine Constraint instances, "
                "got: %s and %s" % (type(lhs), type(rhs))
            )
        self.lhs = lhs_constraint
        self.rhs = rhs_constraint
        self.operator = operator

    def _coordless_match(self, cube):
        return self.operator(
            self.lhs._coordless_match(cube), self.rhs._coordless_match(cube)
        )

    def __repr__(self):
        return "ConstraintCombination(%r, %r, %r)" % (
            self.lhs,
            self.rhs,
            self.operator,
        )

    def _CIM_extract(self, cube):
        return self.operator(
            self.lhs._CIM_extract(cube), self.rhs._CIM_extract(cube)
        )


class _CoordConstraint:
    """Represents the atomic elements which might build up a Constraint."""

    def __init__(self, coord_name, coord_thing):
        """
        Create a coordinate constraint given the coordinate name and a
        thing to compare it with.

        Arguments:

        * coord_name  -  string
            The name of the coordinate to constrain
        * coord_thing
            The object to compare

        """
        self.coord_name = coord_name
        self._coord_thing = coord_thing

    def __repr__(self):
        return "_CoordConstraint(%r, %r)" % (
            self.coord_name,
            self._coord_thing,
        )

    def extract(self, cube):
        """
        Returns the the column based indices of the given cube which
        match the constraint.

        """
        # Cater for scalar cubes by setting the dimensionality to 1
        # when cube.ndim is 0.
        cube_cim = _ColumnIndexManager(cube.ndim or 1)
        try:
            coord = cube.coord(self.coord_name)
        except iris.exceptions.CoordinateNotFoundError:
            cube_cim.all_false()
            return cube_cim
        dims = cube.coord_dims(coord)
        if len(dims) > 1:
            msg = "Cannot apply constraints to multidimensional coordinates"
            raise iris.exceptions.CoordinateMultiDimError(msg)

        try_quick = False
        if callable(self._coord_thing):
            call_func = self._coord_thing
        elif isinstance(self._coord_thing, Iterable) and not isinstance(
            self._coord_thing, (str, iris.coords.Cell)
        ):
            desired_values = list(self._coord_thing)
            # A dramatic speedup can be had if we don't have bounds.
            if coord.has_bounds():

                def call_func(cell):
                    return cell in desired_values

            else:

                def call_func(cell):
                    return cell.point in desired_values

        else:

            def call_func(c):
                return c == self._coord_thing

            try_quick = isinstance(
                coord, iris.coords.DimCoord
            ) and not isinstance(self._coord_thing, iris.coords.Cell)

        # Simple, yet dramatic, optimisation for the monotonic case.
        if try_quick:
            try:
                i = coord.nearest_neighbour_index(self._coord_thing)
            except TypeError:
                try_quick = False
        if try_quick:
            r = np.zeros(coord.shape, dtype=np.bool)
            if coord.cell(i) == self._coord_thing:
                r[i] = True
        else:
            r = np.array([call_func(cell) for cell in coord.cells()])
        if dims:
            cube_cim[dims[0]] = r
        elif not all(r):
            cube_cim.all_false()
        return cube_cim


class _ColumnIndexManager:
    """
    A class to represent column aligned slices which can be operated on
    using ``&``, ``|`` or ``^``.

    ::

        # 4 Dimensional slices
        import numpy as np
        cim = _ColumnIndexManager(4)
        cim[1] = np.array([3, 4, 5]) > 3
        print(cim.as_slice())

    """

    def __init__(self, ndims):
        """
        A _ColumnIndexManager is always created to span the given
        number of dimensions.

        """
        self._column_arrays = [True] * ndims
        self.ndims = ndims

    def __and__(self, other):
        return self._bitwise_operator(other, operator.__and__)

    def __or__(self, other):
        return self._bitwise_operator(other, operator.__or__)

    def __xor__(self, other):
        return self._bitwise_operator(other, operator.__xor__)

    def _bitwise_operator(self, other, operator):
        if not isinstance(other, _ColumnIndexManager):
            return NotImplemented

        if self.ndims != other.ndims:
            raise ValueError(
                "Cannot do %s for %r and %r as they have a "
                "different number of dimensions." % operator
            )
        r = _ColumnIndexManager(self.ndims)
        # iterate over each dimension an combine appropriately
        for i, (lhs, rhs) in enumerate(zip(self, other)):
            r[i] = operator(lhs, rhs)
        return r

    def all_false(self):
        """Turn all slices into False."""
        for i in range(self.ndims):
            self[i] = False

    def __getitem__(self, key):
        return self._column_arrays[key]

    def __setitem__(self, key, value):
        is_vector = isinstance(value, np.ndarray) and value.ndim == 1
        if is_vector or isinstance(value, bool):
            self._column_arrays[key] = value
        else:
            raise TypeError(
                "Expecting value to be a 1 dimensional numpy array"
                ", or a boolean. Got %s" % (type(value))
            )

    def as_slice(self):
        """
        Turns a _ColumnIndexManager into a tuple which can be used in an
        indexing operation.

        If no index is possible, None will be returned.
        """
        result = [None] * self.ndims

        for dim, dimension_array in enumerate(self):
            # If dimension_array has not been set, span the entire dimension
            if isinstance(dimension_array, np.ndarray):
                where_true = np.where(dimension_array)[0]
                # If the array had no True values in it, then the dimension
                # is equivalent to False
                if len(where_true) == 0:
                    result = None
                    break

                # If there was exactly one match, the key should be an integer
                if where_true.shape == (1,):
                    result[dim] = where_true[0]
                else:
                    # Finally, we can either provide a slice if possible,
                    # or a tuple of indices which match. In order to determine
                    # if we can provide a slice, calculate the deltas between
                    # the indices and check if they are the same.
                    delta = np.diff(where_true, axis=0)
                    # if the diff is consistent we can create a slice object
                    if all(delta[0] == delta):
                        result[dim] = slice(
                            where_true[0], where_true[-1] + 1, delta[0]
                        )
                    else:
                        # otherwise, key is a tuple
                        result[dim] = tuple(where_true)

            # Handle the case where dimension_array is a boolean
            elif dimension_array:
                result[dim] = slice(None, None)
            else:
                result = None
                break

        if result is None:
            return result
        else:
            return tuple(result)


def list_of_constraints(constraints):
    """
    Turns the given constraints into a list of valid constraints
    using :func:`as_constraint`.

    """
    if isinstance(constraints, str) or not isinstance(constraints, Iterable):
        constraints = [constraints]

    return [as_constraint(constraint) for constraint in constraints]


def as_constraint(thing):
    """
    Casts an object into a cube constraint where possible, otherwise
    a TypeError will be raised.

    If the given object is already a valid constraint then the given object
    will be returned, else a TypeError will be raised.

    """
    if isinstance(thing, Constraint):
        return thing
    elif thing is None:
        return Constraint()
    elif isinstance(thing, str):
        return Constraint(thing)
    else:
        raise TypeError("%r cannot be cast to a constraint." % thing)


class AttributeConstraint(Constraint):
    """Provides a simple Cube-attribute based :class:`Constraint`."""

    def __init__(self, **attributes):
        """
        Example usage::

            iris.AttributeConstraint(STASH='m01s16i004')

            iris.AttributeConstraint(
                STASH=lambda stash: str(stash).endswith('i005'))

        .. note:: Attribute constraint names are case sensitive.

        """
        self._attributes = attributes
        super().__init__(cube_func=self._cube_func)

    def _cube_func(self, cube):
        match = True
        for name, value in self._attributes.items():
            if name in cube.attributes:
                cube_attr = cube.attributes.get(name)
                # if we have a callable, then call it with the value,
                # otherwise, assert equality
                if callable(value):
                    if not value(cube_attr):
                        match = False
                        break
                else:
                    if cube_attr != value:
                        match = False
                        break
            else:
                match = False
                break
        return match

    def __repr__(self):
        return "AttributeConstraint(%r)" % self._attributes


class NameConstraint(Constraint):
    """Provides a simple Cube name based :class:`Constraint`."""

    def __init__(
        self,
        standard_name="none",
        long_name="none",
        var_name="none",
        STASH="none",
    ):
        """
        Provides a simple Cube name based :class:`Constraint`, which matches
        against each of the names provided, which may be either standard name,
        long name, NetCDF variable name and/or the STASH from the attributes
        dictionary.

        The name constraint will only succeed if *all* of the provided names
        match.

        Kwargs:

        * standard_name:
            A string or callable representing the standard name to match
            against.
        * long_name:
            A string or callable representing the long name to match against.
        * var_name:
            A string or callable representing the NetCDF variable name to match
            against.
        * STASH:
            A string or callable representing the UM STASH code to match
            against.

        .. note::
            The default value of each of the keyword arguments is the string
            "none", rather than the singleton None, as None may be a legitimate
            value to be matched against e.g., to constrain against all cubes
            where the standard_name is not set, then use standard_name=None.

        Returns:

        * Boolean

        Example usage::

            iris.NameConstraint(long_name='air temp', var_name=None)

            iris.NameConstraint(long_name=lambda name: 'temp' in name)

            iris.NameConstraint(standard_name='air_temperature',
                                STASH=lambda stash: stash.item == 203)
        """

        self.standard_name = standard_name
        self.long_name = long_name
        self.var_name = var_name
        self.STASH = STASH
        self._names = ("standard_name", "long_name", "var_name", "STASH")
        super().__init__(cube_func=self._cube_func)

    def _cube_func(self, cube):
        def matcher(target, value):
            if callable(value):
                result = False
                if target is not None:
                    #
                    # Don't pass None through into the callable. Users should
                    # use the "name=None" pattern instead. Otherwise, users
                    # will need to explicitly handle the None case, which is
                    # unnecessary and pretty darn ugly e.g.,
                    #
                    # lambda name: name is not None and name.startswith('ick')
                    #
                    result = value(target)
            else:
                result = value == target
            return result

        match = True
        for name in self._names:
            expected = getattr(self, name)
            if expected != "none":
                if name == "STASH":
                    actual = cube.attributes.get(name)
                else:
                    actual = getattr(cube, name)
                match = matcher(actual, expected)
                # Make this is a short-circuit match.
                if match is False:
                    break

        return match

    def __repr__(self):
        names = []
        for name in self._names:
            value = getattr(self, name)
            if value != "none":
                names.append("{}={!r}".format(name, value))
        return "{}({})".format(self.__class__.__name__, ", ".join(names))



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coord_systems.GeogCS` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.coord_systems import GeogCS


class Test_init_defaults(tests.IrisTest):
    # NOTE: most of the testing for GeogCS is in the legacy test module
    # 'iris.tests.test_coordsystem'.
    # This class *only* tests the defaults for optional constructor args.

    def test_set_optional_args(self):
        # Check that setting the optional (non-ellipse) argument works.
        crs = GeogCS(1.0, longitude_of_prime_meridian=-85)
        self.assertEqualAndKind(crs.longitude_of_prime_meridian, -85.0)

    def _check_crs_defaults(self, crs):
        # Check for property defaults when no kwargs options were set.
        # NOTE: except ellipsoid, which is done elsewhere.
        radius = float(crs.semi_major_axis)
        self.assertEqualAndKind(crs.semi_major_axis, radius)  # just the kind
        self.assertEqualAndKind(crs.semi_minor_axis, radius)
        self.assertEqualAndKind(crs.inverse_flattening, 0.0)
        self.assertEqualAndKind(crs.longitude_of_prime_meridian, 0.0)

    def test_no_optional_args(self):
        # Check expected properties with no optional args.
        crs = GeogCS(1.0)
        self._check_crs_defaults(crs)

    def test_optional_args_None(self):
        # Check expected properties with optional args=None.
        crs = GeogCS(
            1.0,
            semi_minor_axis=None,
            inverse_flattening=None,
            longitude_of_prime_meridian=None,
        )
        self._check_crs_defaults(crs)


if __name__ == "__main__":
    tests.main()



import matplotlib.pyplot as plt

import iris
import iris.quickplot as qplt


fname = iris.sample_data_path("air_temp.pp")
temperature_cube = iris.load_cube(fname)

# Add a contour, and put the result in a variable called contour.
contour = qplt.contour(temperature_cube)

# Add coastlines to the map created by contour.
plt.gca().coastlines()

# Add contour labels based on the contour we have just created.
plt.clabel(contour, inline=False)

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.VARIANCE` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np
import numpy.ma as ma

from iris._lazy_data import as_lazy_data, as_concrete_data
from iris.analysis import VARIANCE
import iris.cube
from iris.coords import DimCoord


class Test_units_func(tests.IrisTest):
    def test(self):
        self.assertIsNotNone(VARIANCE.units_func)
        mul = mock.Mock(return_value=mock.sentinel.new_unit)
        units = mock.Mock(__mul__=mul)
        new_units = VARIANCE.units_func(units)
        # Make sure the VARIANCE units_func tries to square the units.
        mul.assert_called_once_with(units)
        self.assertEqual(new_units, mock.sentinel.new_unit)


class Test_masked(tests.IrisTest):
    def setUp(self):
        self.cube = iris.cube.Cube(ma.masked_equal([1, 2, 3, 4, 5], 3))
        self.cube.add_dim_coord(DimCoord([6, 7, 8, 9, 10], long_name="foo"), 0)

    def test_ma_ddof0(self):
        cube = self.cube.collapsed("foo", VARIANCE, ddof=0)
        expected = 10 / 4.0
        self.assertArrayEqual(np.var(self.cube.data, ddof=0), expected)
        self.assertArrayAlmostEqual(cube.data, expected)

    def test_ma_ddof1(self):
        cube = self.cube.collapsed("foo", VARIANCE, ddof=1)
        expected = 10 / 3.0
        self.assertArrayEqual(np.var(self.cube.data, ddof=1), expected)
        self.assertArrayEqual(cube.data, expected)

        # test that the default ddof is 1
        default_cube = self.cube.collapsed("foo", VARIANCE)
        self.assertArrayEqual(cube.data, default_cube.data)


class Test_lazy_aggregate(tests.IrisTest):
    def test_ddof_one(self):
        array = as_lazy_data(np.arange(8))
        var = VARIANCE.lazy_aggregate(array, axis=0, ddof=1)
        result = as_concrete_data(var)
        self.assertArrayAlmostEqual(result, np.array(6.0))

    def test_ddof_zero(self):
        array = as_lazy_data(np.arange(8))
        var = VARIANCE.lazy_aggregate(array, axis=0, ddof=0)
        result = as_concrete_data(var)
        self.assertArrayAlmostEqual(result, np.array(5.25))


class Test_name(tests.IrisTest):
    def test(self):
        self.assertEqual(VARIANCE.name(), "variance")


class Test_aggregate_shape(tests.IrisTest):
    def test(self):
        shape = ()
        kwargs = dict()
        self.assertTupleEqual(VARIANCE.aggregate_shape(**kwargs), shape)
        kwargs = dict(bat="man", wonder="woman")
        self.assertTupleEqual(VARIANCE.aggregate_shape(**kwargs), shape)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test pickling of Iris objects.

"""

# Import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

import io
import pickle

import cf_units

import iris
from iris._lazy_data import as_concrete_data


class TestPickle(tests.IrisTest):
    def pickle_then_unpickle(self, obj):
        """
        Returns a generator of ("pickle protocol number", object) tuples.

        """
        for protocol in range(1 + pickle.HIGHEST_PROTOCOL):
            bio = io.BytesIO()
            pickle.dump(obj, bio, protocol)

            # move the bio back to the start and reconstruct
            bio.seek(0)
            reconstructed_obj = pickle.load(bio)

            yield protocol, reconstructed_obj

    @staticmethod
    def _real_data(cube):
        # Get the concrete data of the cube for performing data values
        # comparison checks.
        return as_concrete_data(cube.core_data())

    def assertCubeData(self, cube1, cube2):
        self.assertArrayEqual(self._real_data(cube1), self._real_data(cube2))

    @tests.skip_data
    def test_cube_pickle(self):
        cube = iris.load_cube(
            tests.get_data_path(("PP", "globClim1", "theta.pp"))
        )
        self.assertTrue(cube.has_lazy_data())
        self.assertCML(
            cube, ("cube_io", "pickling", "theta.cml"), checksum=False
        )

        for p, recon_cube in self.pickle_then_unpickle(cube):
            self.assertTrue(recon_cube.has_lazy_data())
            self.assertCML(
                recon_cube,
                ("cube_io", "pickling", "theta.cml"),
                checksum=False,
            )
            self.assertCubeData(cube, recon_cube)

    @tests.skip_data
    def test_cube_with_coord_points(self):
        filename = tests.get_data_path(
            ("NetCDF", "rotated", "xy", "rotPole_landAreaFraction.nc")
        )
        cube = iris.load_cube(filename)
        # Pickle and unpickle. Do not perform any CML tests
        # to avoid side effects.
        _, recon_cube = next(self.pickle_then_unpickle(cube))
        self.assertEqual(recon_cube, cube)

    @tests.skip_data
    def test_cubelist_pickle(self):
        cubelist = iris.load(
            tests.get_data_path(("PP", "COLPEX", "theta_and_orog_subset.pp"))
        )
        single_cube = cubelist[0]

        self.assertCML(cubelist, ("cube_io", "pickling", "cubelist.cml"))
        self.assertCML(single_cube, ("cube_io", "pickling", "single_cube.cml"))

        for _, reconstructed_cubelist in self.pickle_then_unpickle(cubelist):
            self.assertCML(
                reconstructed_cubelist, ("cube_io", "pickling", "cubelist.cml")
            )
            self.assertCML(
                reconstructed_cubelist[0],
                ("cube_io", "pickling", "single_cube.cml"),
            )

            for cube_orig, cube_reconstruct in zip(
                cubelist, reconstructed_cubelist
            ):
                self.assertArrayEqual(cube_orig.data, cube_reconstruct.data)
                self.assertEqual(cube_orig, cube_reconstruct)

    def test_picking_equality_misc(self):
        items_to_test = [
            cf_units.Unit(
                "hours since 2007-01-15 12:06:00",
                calendar=cf_units.CALENDAR_STANDARD,
            ),
            cf_units.as_unit("1"),
            cf_units.as_unit("meters"),
            cf_units.as_unit("no-unit"),
            cf_units.as_unit("unknown"),
        ]

        for orig_item in items_to_test:
            for protocol, reconst_item in self.pickle_then_unpickle(orig_item):
                fail_msg = (
                    "Items are different after pickling "
                    "at protocol {}.\nOrig item: {!r}\nNew item: {!r}"
                )
                fail_msg = fail_msg.format(protocol, orig_item, reconst_item)
                self.assertEqual(orig_item, reconst_item, fail_msg)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import os

import iris


@tests.skip_data
class TestCallbacks(tests.IrisTest):
    def test_pp_callback(self):
        def pp_callback(cube, field, filename):
            cube.attributes["filename"] = os.path.basename(filename)
            cube.attributes["lbyr"] = field.lbyr

        fname = tests.get_data_path(("PP", "aPPglob1", "global.pp"))
        cube = iris.load_cube(fname, callback=pp_callback)
        self.assertCML(cube, ["uri_callback", "pp_global.cml"])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the hybrid vertical coordinate representations.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import warnings

import numpy as np

from iris.aux_factory import HybridHeightFactory, HybridPressureFactory
import iris
import iris.tests.stock


@tests.skip_plot
@tests.skip_data
class TestRealistic4d(tests.GraphicsTest):
    def setUp(self):
        super().setUp()
        self.cube = iris.tests.stock.realistic_4d()
        self.altitude = self.cube.coord("altitude")

    def test_metadata(self):
        self.assertEqual(self.altitude.units, "m")
        self.assertIsNone(self.altitude.coord_system)
        self.assertEqual(self.altitude.attributes, {"positive": "up"})

    def test_points(self):
        self.assertAlmostEqual(
            self.altitude.points.min(), np.float32(191.84892)
        )
        self.assertAlmostEqual(self.altitude.points.max(), np.float32(40000))

    def test_transpose(self):
        self.assertCML(self.cube, ("stock", "realistic_4d.cml"))
        self.cube.transpose()
        self.assertCML(self.cube, ("derived", "transposed.cml"))

    def test_indexing(self):
        cube = self.cube[:, :, 0, 0]
        # Make sure the derived 'altitude' coordinate survived the indexing.
        _ = cube.coord("altitude")
        self.assertCML(cube, ("derived", "column.cml"))

    def test_removing_derived_coord(self):
        cube = self.cube
        cube.remove_coord("altitude")
        self.assertCML(cube, ("derived", "removed_derived_coord.cml"))

    def test_removing_sigma(self):
        # Check the cube remains OK when sigma is removed.
        cube = self.cube
        cube.remove_coord("sigma")
        self.assertCML(cube, ("derived", "removed_sigma.cml"))
        self.assertString(str(cube), ("derived", "removed_sigma.__str__.txt"))

        # Check the factory now only has surface_altitude and delta dependencies.
        factory = cube.aux_factory(name="altitude")
        t = [
            key
            for key, coord in factory.dependencies.items()
            if coord is not None
        ]
        self.assertCountEqual(t, ["orography", "delta"])

    def test_removing_orography(self):
        # Check the cube remains OK when the orography is removed.
        cube = self.cube
        cube.remove_coord("surface_altitude")
        self.assertCML(cube, ("derived", "removed_orog.cml"))
        self.assertString(str(cube), ("derived", "removed_orog.__str__.txt"))

        # Check the factory now only has sigma and delta dependencies.
        factory = cube.aux_factory(name="altitude")
        t = [
            key
            for key, coord in factory.dependencies.items()
            if coord is not None
        ]
        self.assertCountEqual(t, ["sigma", "delta"])

    def test_derived_coords(self):
        derived_coords = self.cube.derived_coords
        self.assertEqual(len(derived_coords), 1)
        altitude = derived_coords[0]
        self.assertEqual(altitude.standard_name, "altitude")
        self.assertEqual(altitude.attributes, {"positive": "up"})

    def test_aux_factory(self):
        factory = self.cube.aux_factory(name="altitude")
        self.assertEqual(factory.standard_name, "altitude")
        self.assertEqual(factory.attributes, {"positive": "up"})

    def test_aux_factory_var_name(self):
        factory = self.cube.aux_factory(name="altitude")
        factory.var_name = "alt"
        factory = self.cube.aux_factory(var_name="alt")
        self.assertEqual(factory.standard_name, "altitude")
        self.assertEqual(factory.attributes, {"positive": "up"})

    def test_no_orography(self):
        # Get rid of the normal hybrid-height factory.
        cube = self.cube
        factory = cube.aux_factory(name="altitude")
        cube.remove_aux_factory(factory)

        # Add a new one which only references level_height & sigma.
        delta = cube.coord("level_height")
        sigma = cube.coord("sigma")
        factory = HybridHeightFactory(delta, sigma)
        cube.add_aux_factory(factory)

        self.assertEqual(len(cube.aux_factories), 1)
        self.assertEqual(len(cube.derived_coords), 1)
        self.assertString(str(cube), ("derived", "no_orog.__str__.txt"))
        self.assertCML(cube, ("derived", "no_orog.cml"))

    def test_invalid_dependencies(self):
        # Must have either delta or orography
        with self.assertRaises(ValueError):
            _ = HybridHeightFactory()
        sigma = self.cube.coord("sigma")
        with self.assertRaises(ValueError):
            _ = HybridHeightFactory(sigma=sigma)

        # Orography must not have bounds
        with warnings.catch_warnings():
            # Cause all warnings to raise Exceptions
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                _ = HybridHeightFactory(orography=sigma)

    def test_bounded_orography(self):
        # Start with everything normal
        orog = self.cube.coord("surface_altitude")
        altitude = self.cube.coord("altitude")
        self.assertIsInstance(altitude.bounds, np.ndarray)

        # Make sure altitude still works OK if orography was messed
        # with *after* altitude was created.
        orog.bounds = np.zeros(orog.shape + (4,))

        # Check that altitude derivation now issues a warning.
        msg = "Orography.* bounds.* being disregarded"
        with warnings.catch_warnings():
            # Cause all warnings to raise Exceptions
            warnings.simplefilter("error")
            with self.assertRaisesRegex(UserWarning, msg):
                self.cube.coord("altitude")


@tests.skip_data
class TestHybridPressure(tests.IrisTest):
    def setUp(self):
        # Convert the hybrid-height into hybrid-pressure...
        cube = iris.tests.stock.realistic_4d()

        # Get rid of the normal hybrid-height factory.
        factory = cube.aux_factory(name="altitude")
        cube.remove_aux_factory(factory)

        # Mangle the height coords into pressure coords.
        delta = cube.coord("level_height")
        delta.rename("level_pressure")
        delta.units = "Pa"
        sigma = cube.coord("sigma")
        ref = cube.coord("surface_altitude")
        ref.rename("surface_air_pressure")
        ref.units = "Pa"

        factory = HybridPressureFactory(delta, sigma, ref)
        cube.add_aux_factory(factory)
        self.cube = cube
        self.air_pressure = self.cube.coord("air_pressure")

    def test_metadata(self):
        self.assertEqual(self.air_pressure.units, "Pa")
        self.assertIsNone(self.air_pressure.coord_system)
        self.assertEqual(self.air_pressure.attributes, {})

    def test_points(self):
        points = self.air_pressure.points
        self.assertEqual(points.dtype, np.float32)
        self.assertAlmostEqual(points.min(), np.float32(191.84892))
        self.assertAlmostEqual(points.max(), np.float32(40000))

        # Convert the reference surface to float64 and check the
        # derived coordinate becomes float64.
        temp = self.cube.coord("surface_air_pressure").points
        temp = temp.astype("f8")
        self.cube.coord("surface_air_pressure").points = temp
        points = self.cube.coord("air_pressure").points
        self.assertEqual(points.dtype, np.float64)
        self.assertAlmostEqual(points.min(), 191.8489257)
        self.assertAlmostEqual(points.max(), 40000)

    def test_invalid_dependencies(self):
        # Must have either delta or surface_air_pressure
        with self.assertRaises(ValueError):
            _ = HybridPressureFactory()
        sigma = self.cube.coord("sigma")
        with self.assertRaises(ValueError):
            _ = HybridPressureFactory(sigma=sigma)

        # Surface pressure must not have bounds
        with warnings.catch_warnings():
            # Cause all warnings to raise Exceptions
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                _ = HybridPressureFactory(
                    sigma=sigma, surface_air_pressure=sigma
                )

    def test_bounded_surface_pressure(self):
        # Start with everything normal
        surface_pressure = self.cube.coord("surface_air_pressure")
        pressure = self.cube.coord("air_pressure")
        self.assertIsInstance(pressure.bounds, np.ndarray)

        # Make sure pressure still works OK if surface pressure was messed
        # with *after* pressure was created.
        surface_pressure.bounds = np.zeros(surface_pressure.shape + (4,))

        # Check that air_pressure derivation now issues a warning.
        msg = "Surface pressure.* bounds.* being disregarded"
        with warnings.catch_warnings():
            # Cause all warnings to raise Exceptions
            warnings.simplefilter("error")
            with self.assertRaisesRegex(UserWarning, msg):
                self.cube.coord("air_pressure")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.quickplot.scatter` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests
from iris.tests.unit.plot import TestGraphicStringCoord

if tests.MPL_AVAILABLE:
    import iris.quickplot as qplt


@tests.skip_plot
class TestStringCoordPlot(TestGraphicStringCoord):
    def setUp(self):
        super().setUp()
        self.cube = self.cube[0, :]

    def test_xaxis_labels(self):
        qplt.scatter(self.cube.coord("str_coord"), self.cube)
        self.assertBoundsTickLabels("xaxis")

    def test_yaxis_labels(self):
        qplt.scatter(self.cube, self.cube.coord("str_coord"))
        self.assertBoundsTickLabels("yaxis")


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the coordinate categorisation functions.
"""

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import warnings

import cf_units
import numpy as np

import iris
import iris.coord_categorisation as ccat


CATEGORISATION_FUNCS = (
    ccat.add_day_of_month,
    ccat.add_day_of_year,
    ccat.add_weekday,
    ccat.add_weekday_fullname,
    ccat.add_weekday_number,
    ccat.add_month,
    ccat.add_month_fullname,
    ccat.add_month_number,
    ccat.add_year,
    ccat.add_season,
    ccat.add_season_number,
    ccat.add_season_year,
    ccat.add_season_membership,
)


class TestCategorisations(tests.IrisTest):
    def setUp(self):
        # make a series of 'day numbers' for the time, that slide across month
        # boundaries
        day_numbers = np.arange(0, 600, 27, dtype=np.int32)

        cube = iris.cube.Cube(
            day_numbers, long_name="test cube", units="metres"
        )

        # use day numbers as data values also (don't actually use this for
        # anything)
        cube.data = day_numbers

        time_coord = iris.coords.DimCoord(
            day_numbers,
            standard_name="time",
            units=cf_units.Unit("days since epoch", "gregorian"),
        )
        cube.add_dim_coord(time_coord, 0)

        self.cube = cube
        self.time_coord = time_coord

    def test_bad_coord(self):
        for func in CATEGORISATION_FUNCS:
            kwargs = {"name": "my_category"}
            if func is ccat.add_season_membership:
                kwargs["season"] = "djf"
            with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
                func(self.cube, "DOES NOT EXIST", **kwargs)

    def test_explicit_result_names(self):
        result_name = "my_category"
        fmt = "Missing/incorrectly named result for {0!r}"
        for func in CATEGORISATION_FUNCS:
            # Specify source coordinate by name
            cube = self.cube.copy()
            kwargs = {"name": result_name}
            if func is ccat.add_season_membership:
                kwargs["season"] = "djf"
            with warnings.catch_warnings(record=True):
                func(cube, "time", **kwargs)
            result_coords = cube.coords(result_name)
            self.assertEqual(len(result_coords), 1, fmt.format(func.__name__))
            # Specify source coordinate by coordinate reference
            cube = self.cube.copy()
            time = cube.coord("time")
            with warnings.catch_warnings(record=True):
                func(cube, time, **kwargs)
            result_coords = cube.coords(result_name)
            self.assertEqual(len(result_coords), 1, fmt.format(func.__name__))

    def test_basic(self):
        cube = self.cube
        time_coord = self.time_coord

        ccat.add_year(cube, time_coord, "my_year")
        ccat.add_day_of_month(cube, time_coord, "my_day_of_month")
        ccat.add_day_of_year(cube, time_coord, "my_day_of_year")

        ccat.add_month(cube, time_coord, "my_month")
        ccat.add_month_fullname(cube, time_coord, "my_month_fullname")
        ccat.add_month_number(cube, time_coord, "my_month_number")

        ccat.add_weekday(cube, time_coord, "my_weekday")
        ccat.add_weekday_number(cube, time_coord, "my_weekday_number")
        ccat.add_weekday_fullname(cube, time_coord, "my_weekday_fullname")

        ccat.add_season(cube, time_coord, "my_season")
        ccat.add_season_number(cube, time_coord, "my_season_number")
        ccat.add_season_year(cube, time_coord, "my_season_year")

        # also test 'generic' categorisation interface
        def _month_in_quarter(coord, pt_value):
            date = coord.units.num2date(pt_value)
            return (date.month - 1) % 3

        ccat.add_categorised_coord(
            cube, "my_month_in_quarter", time_coord, _month_in_quarter
        )

        # To ensure consistent results between 32-bit and 64-bit
        # platforms, ensure all the numeric categorisation coordinates
        # are always stored as int64.
        for coord in cube.coords():
            if coord.long_name is not None and coord.points.dtype.kind == "i":
                coord.points = coord.points.astype(np.int64)

        # check values
        self.assertCML(cube, ("categorisation", "quickcheck.cml"))

    def test_add_season_nonstandard(self):
        # season categorisations work for non-standard seasons?
        cube = self.cube
        time_coord = self.time_coord
        seasons = ["djfm", "amjj", "ason"]
        ccat.add_season(cube, time_coord, name="seasons", seasons=seasons)
        ccat.add_season_number(
            cube, time_coord, name="season_numbers", seasons=seasons
        )
        ccat.add_season_year(
            cube, time_coord, name="season_years", seasons=seasons
        )
        self.assertCML(cube, ("categorisation", "customcheck.cml"))

    def test_add_season_membership(self):
        # season membership identifies correct seasons?
        season = "djf"
        ccat.add_season_membership(self.cube, "time", season, name="in_season")
        ccat.add_season(self.cube, "time")
        coord_season = self.cube.coord("season")
        coord_membership = self.cube.coord("in_season")
        season_locations = np.where(coord_season.points == season)[0]
        membership_locations = np.where(coord_membership.points)[0]
        self.assertArrayEqual(membership_locations, season_locations)

    def test_add_season_invalid_spec(self):
        # custom seasons with an invalid season raises an error?
        seasons = ("djf", "maj", "jja", "son")  # MAJ not a season!
        for func in (
            ccat.add_season,
            ccat.add_season_year,
            ccat.add_season_number,
        ):
            with self.assertRaises(ValueError):
                func(self.cube, "time", name="my_category", seasons=seasons)

    def test_add_season_repeated_months(self):
        # custom seasons with repeated months raises an error?
        seasons = ("djfm", "mam", "jja", "son")
        for func in (
            ccat.add_season,
            ccat.add_season_year,
            ccat.add_season_number,
        ):
            with self.assertRaises(ValueError):
                func(self.cube, "time", name="my_category", seasons=seasons)

    def test_add_season_missing_months(self):
        # custom seasons with missing months raises an error?
        seasons = ("djfm", "amjj")
        for func in (
            ccat.add_season,
            ccat.add_season_year,
            ccat.add_season_number,
        ):
            with self.assertRaises(ValueError):
                func(self.cube, "time", name="my_category", seasons=seasons)

    def test_add_season_membership_invalid_spec(self):
        season = "maj"  # not a season!
        with self.assertRaises(ValueError):
            ccat.add_season_membership(
                self.cube, "time", season, name="maj_season"
            )


if __name__ == "__main__":
    tests.main()
